# Stubs for networkx.algorithms.cycles (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def cycle_basis(G, root: Optional[Any] = ...): ...
def simple_cycles(G): ...
def recursive_simple_cycles(G): ...
def find_cycle(G, source: Optional[Any] = ..., orientation: str = ...): ...
def minimum_cycle_basis(G, weight: Optional[Any] = ...): ...



# Stubs for networkx.linalg (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.linalg.algebraicconnectivity import *
from networkx.linalg.attrmatrix import *
from networkx.linalg.graphmatrix import *
from networkx.linalg.laplacianmatrix import *
from networkx.linalg.modularitymatrix import *
from networkx.linalg.spectrum import *



# Stubs for networkx.generators.random_graphs (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def fast_gnp_random_graph(n, p, seed: Optional[Any] = ..., directed: bool = ...): ...
def gnp_random_graph(n, p, seed: Optional[Any] = ..., directed: bool = ...): ...

binomial_graph = gnp_random_graph
erdos_renyi_graph = gnp_random_graph

def dense_gnm_random_graph(n, m, seed: Optional[Any] = ...): ...
def gnm_random_graph(n, m, seed: Optional[Any] = ..., directed: bool = ...): ...
def newman_watts_strogatz_graph(n, k, p, seed: Optional[Any] = ...): ...
def watts_strogatz_graph(n, k, p, seed: Optional[Any] = ...): ...
def connected_watts_strogatz_graph(n, k, p, tries: int = ..., seed: Optional[Any] = ...): ...
def random_regular_graph(d, n, seed: Optional[Any] = ...): ...
def barabasi_albert_graph(n, m, seed: Optional[Any] = ...): ...
def extended_barabasi_albert_graph(n, m, p, q, seed: Optional[Any] = ...): ...
def powerlaw_cluster_graph(n, m, p, seed: Optional[Any] = ...): ...
def random_lobster(n, p1, p2, seed: Optional[Any] = ...): ...
def random_shell_graph(constructor, seed: Optional[Any] = ...): ...
def random_powerlaw_tree(n, gamma: int = ..., seed: Optional[Any] = ..., tries: int = ...): ...
def random_powerlaw_tree_sequence(
    n, gamma: int = ..., seed: Optional[Any] = ..., tries: int = ...
): ...
def random_kernel_graph(
    n, kernel_integral, kernel_root: Optional[Any] = ..., seed: Optional[Any] = ...
): ...



# Stubs for networkx.algorithms.approximation.clique (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def max_clique(G): ...
def clique_removal(G): ...
def large_clique_size(G): ...



# Stubs for networkx.algorithms.centrality.closeness (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def closeness_centrality(
    G,
    u: Optional[Any] = ...,
    distance: Optional[Any] = ...,
    wf_improved: bool = ...,
    reverse: bool = ...,
): ...



# Stubs for networkx.algorithms.assortativity.pairs (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def node_attribute_xy(G, attribute, nodes: Optional[Any] = ...): ...
def node_degree_xy(
    G,
    x: str = ...,
    y: str = ...,
    weight: Optional[Any] = ...,
    nodes: Optional[Any] = ...,
): ...



# Stubs for networkx.drawing.nx_agraph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def from_agraph(A, create_using: Optional[Any] = ...): ...
def to_agraph(N): ...
def write_dot(G, path): ...
def read_dot(path): ...
def graphviz_layout(G, prog: str = ..., root: Optional[Any] = ..., args: str = ...): ...
def pygraphviz_layout(G, prog: str = ..., root: Optional[Any] = ..., args: str = ...): ...
def view_pygraphviz(
    G,
    edgelabel: Optional[Any] = ...,
    prog: str = ...,
    args: str = ...,
    suffix: str = ...,
    path: Optional[Any] = ...,
): ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class PROV(DefinedNamespace):
    activityOfInfluence: URIRef
    agentOfInfluence: URIRef
    contributed: URIRef
    ended: URIRef
    entityOfInfluence: URIRef
    generalizationOf: URIRef
    generatedAsDerivation: URIRef
    hadDelegate: URIRef
    hadDerivation: URIRef
    hadInfluence: URIRef
    hadRevision: URIRef
    informed: URIRef
    locationOf: URIRef
    qualifiedAssociationOf: URIRef
    qualifiedAttributionOf: URIRef
    qualifiedCommunicationOf: URIRef
    qualifiedDelegationOf: URIRef
    qualifiedDerivationOf: URIRef
    qualifiedEndOf: URIRef
    qualifiedGenerationOf: URIRef
    qualifiedInfluenceOf: URIRef
    qualifiedInvalidationOf: URIRef
    qualifiedQuotationOf: URIRef
    qualifiedSourceOf: URIRef
    qualifiedStartOf: URIRef
    qualifiedUsingActivity: URIRef
    quotedAs: URIRef
    revisedEntity: URIRef
    started: URIRef
    wasActivityOfInfluence: URIRef
    wasAssociateFor: URIRef
    wasMemberOf: URIRef
    wasPlanOf: URIRef
    wasPrimarySourceOf: URIRef
    wasRoleIn: URIRef
    wasUsedBy: URIRef
    wasUsedInDerivation: URIRef
    aq: URIRef
    category: URIRef
    component: URIRef
    constraints: URIRef
    definition: URIRef
    dm: URIRef
    editorialNote: URIRef
    editorsDefinition: URIRef
    inverse: URIRef
    n: URIRef
    order: URIRef
    qualifiedForm: URIRef
    sharesDefinitionWith: URIRef
    specializationOf: URIRef
    todo: URIRef
    unqualifiedForm: URIRef
    wasRevisionOf: URIRef
    Accept: URIRef
    Activity: URIRef
    ActivityInfluence: URIRef
    Agent: URIRef
    AgentInfluence: URIRef
    Association: URIRef
    Attribution: URIRef
    Bundle: URIRef
    Collection: URIRef
    Communication: URIRef
    Contribute: URIRef
    Contributor: URIRef
    Copyright: URIRef
    Create: URIRef
    Creator: URIRef
    Delegation: URIRef
    Derivation: URIRef
    Dictionary: URIRef
    DirectQueryService: URIRef
    EmptyDictionary: URIRef
    End: URIRef
    Entity: URIRef
    EntityInfluence: URIRef
    Generation: URIRef
    Influence: URIRef
    Insertion: URIRef
    InstantaneousEvent: URIRef
    Invalidation: URIRef
    KeyEntityPair: URIRef
    Location: URIRef
    Modify: URIRef
    Organization: URIRef
    Person: URIRef
    Plan: URIRef
    PrimarySource: URIRef
    Publish: URIRef
    Publisher: URIRef
    Quotation: URIRef
    Removal: URIRef
    Replace: URIRef
    Revision: URIRef
    RightsAssignment: URIRef
    RightsHolder: URIRef
    Role: URIRef
    ServiceDescription: URIRef
    SoftwareAgent: URIRef
    Start: URIRef
    Submit: URIRef
    Usage: URIRef
    atTime: URIRef
    endedAtTime: URIRef
    generatedAtTime: URIRef
    invalidatedAtTime: URIRef
    provenanceUriTemplate: URIRef
    removedKey: URIRef
    startedAtTime: URIRef
    value: URIRef
    pairEntity: URIRef
    pairKey: URIRef
    EmptyCollection: URIRef
    actedOnBehalfOf: URIRef
    activity: URIRef
    agent: URIRef
    alternateOf: URIRef
    asInBundle: URIRef
    atLocation: URIRef
    derivedByInsertionFrom: URIRef
    derivedByRemovalFrom: URIRef
    describesService: URIRef
    dictionary: URIRef
    entity: URIRef
    generated: URIRef
    hadActivity: URIRef
    hadDictionaryMember: URIRef
    hadGeneration: URIRef
    hadMember: URIRef
    hadPlan: URIRef
    hadPrimarySource: URIRef
    hadRole: URIRef
    hadUsage: URIRef
    has_anchor: URIRef
    has_provenance: URIRef
    has_query_service: URIRef
    influenced: URIRef
    influencer: URIRef
    insertedKeyEntityPair: URIRef
    invalidated: URIRef
    mentionOf: URIRef
    pingback: URIRef
    qualifiedAssociation: URIRef
    qualifiedAttribution: URIRef
    qualifiedCommunication: URIRef
    qualifiedDelegation: URIRef
    qualifiedDerivation: URIRef
    qualifiedEnd: URIRef
    qualifiedGeneration: URIRef
    qualifiedInfluence: URIRef
    qualifiedInsertion: URIRef
    qualifiedInvalidation: URIRef
    qualifiedPrimarySource: URIRef
    qualifiedQuotation: URIRef
    qualifiedRemoval: URIRef
    qualifiedRevision: URIRef
    qualifiedStart: URIRef
    qualifiedUsage: URIRef
    used: URIRef
    wasAssociatedWith: URIRef
    wasAttributedTo: URIRef
    wasDerivedFrom: URIRef
    wasEndedBy: URIRef
    wasGeneratedBy: URIRef
    wasInfluencedBy: URIRef
    wasInformedBy: URIRef
    wasInvalidatedBy: URIRef
    wasQuotedFrom: URIRef
    wasStartedBy: URIRef



# Stubs for networkx.readwrite.nx_yaml (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def write_yaml(G_to_be_yaml, path_for_yaml_output, **kwds): ...
def read_yaml(path): ...



from typing import Any, Dict, List, Sequence, Union

PY3: Any
str_type = str
GRAPH_ATTRIBUTES: Any
EDGE_ATTRIBUTES: Any
NODE_ATTRIBUTES: Any
CLUSTER_ATTRIBUTES: Any
DEFAULT_PROGRAMS: Any

def is_windows() -> bool: ...
def is_anaconda() -> bool: ...
def get_executable_extension() -> str: ...
def graph_from_dot_data(s: str) -> List["Dot"]: ...

class Common:
    def set_parent_graph(self, parent_graph: "Graph") -> None: ...
    def get_parent_graph(self) -> "Graph": ...
    def set(self, name: str, value: str) -> None: ...
    def get(self, name: str) -> str: ...
    def get_attributes(self) -> Dict[str, str]: ...
    def set_sequence(self, seq: str) -> None: ...
    def get_sequence(self) -> str: ...
    def create_attribute_methods(self, obj_attributes: List[str]) -> None: ...

class Error(Exception):
    value: Any
    def __init__(self, value: str) -> None: ...

class InvocationException(Exception):
    value: Any
    def __init__(self, value: str) -> None: ...

class Node(Common):
    obj_dict: Any
    def __init__(self, name: str = ..., obj_dict: Any | None = ..., **attrs: str) -> None: ...
    def set_name(self, node_name: str) -> None: ...
    def get_name(self) -> str: ...
    def get_port(self) -> str: ...
    def add_style(self, style: str) -> None: ...
    def to_string(self) -> str: ...

class Edge(Common):
    obj_dict: Any
    def __init__(
        self,
        src: str = ...,
        dst: str = ...,
        obj_dict: Any | None = ...,
        **attrs: Dict[str, str],
    ) -> None: ...
    def get_source(self) -> str: ...
    def get_destination(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, edge: Any) -> bool: ...
    def parse_node_ref(self, node_str: str) -> str: ...
    def to_string(self) -> str: ...

class Graph(Common):
    obj_dict: Any
    def __init__(
        self,
        graph_name: str = ...,
        obj_dict: Any | None = ...,
        graph_type: str = ...,
        strict: bool = ...,
        suppress_disconnected: bool = ...,
        simplify: bool = ...,
        **attrs: Dict[str, str],
    ) -> None: ...
    def get_graph_type(self) -> str: ...
    def get_top_graph_type(self) -> str: ...
    def set_graph_defaults(self, **attrs: Dict[str, str]) -> None: ...
    def get_graph_defaults(self, **attrs: Dict[str, str]) -> Dict[str, str]: ...
    def set_node_defaults(self, **attrs: Dict[str, str]) -> None: ...
    def get_node_defaults(self, **attrs: Dict[str, str]) -> Dict[str, str]: ...
    def set_edge_defaults(self, **attrs: Dict[str, str]) -> None: ...
    def get_edge_defaults(self, **attrs: Dict[str, str]) -> Dict[str, str]: ...
    def set_simplify(self, simplify: bool) -> None: ...
    def get_simplify(self) -> bool: ...
    def set_type(self, graph_type: str) -> None: ...
    def get_type(self) -> str: ...
    def set_name(self, graph_name: str) -> None: ...
    def get_name(self) -> str: ...
    def set_strict(self, val: bool) -> None: ...
    def get_strict(self, val: Any) -> bool: ...
    def set_suppress_disconnected(self, val: bool) -> None: ...
    def get_suppress_disconnected(self, val: Any) -> None: ...
    def get_next_sequence_number(self) -> int: ...
    def add_node(self, graph_node: Node) -> None: ...
    def del_node(self, name: Union[str, Node], index: int | None = ...) -> bool: ...
    def get_node(self, name: str) -> Node: ...
    def get_nodes(self) -> List[Node]: ...
    def get_node_list(self) -> List[Node]: ...
    def add_edge(self, graph_edge: Edge) -> None: ...
    def del_edge(
        self,
        src_or_list: Union[Sequence[Node], Node],
        dst: str | int | None = ...,
        index: int | None = ...,
    ) -> bool: ...
    def get_edge(
        self, src_or_list: Union[Sequence[Node], Node], dst: Any | None = ...
    ) -> List[Edge]: ...
    def get_edges(self) -> List[Edge]: ...
    def get_edge_list(self) -> List[Edge]: ...
    def add_subgraph(self, sgraph: Union["Subgraph", "Cluster"]) -> None: ...
    def get_subgraph(self, name: str) -> List["Subgraph"]: ...
    def get_subgraphs(self) -> List["Subgraph"]: ...
    def get_subgraph_list(self) -> List["Subgraph"]: ...
    def set_parent_graph(self, parent_graph: "Graph") -> None: ...
    def to_string(self) -> str: ...

class Subgraph(Graph):
    def __init__(
        self,
        graph_name: str = ...,
        obj_dict: Any | Dict[str, str] = ...,
        suppress_disconnected: bool = ...,
        simplify: bool = ...,
        **attrs: Dict[str, str],
    ) -> None: ...

class Cluster(Graph):
    def __init__(
        self,
        graph_name: str = ...,
        obj_dict: Any | Dict[str, str] = ...,
        suppress_disconnected: bool = ...,
        simplify: bool = ...,
        **attrs: Dict[str, str],
    ) -> None: ...

class Dot(Graph):
    shape_files: Any
    formats: Any
    prog: str
    def __init__(self, *argsl: Any, **argsd: Dict[str, str]): ...
    def set_shape_files(self, file_paths: Union[str, Sequence[str]]) -> None: ...
    def set_prog(self, prog: str) -> None: ...
    def write(
        self,
        path: str,
        prog: Any | str = ...,
        format: str = ...,
        encoding: Any | str = ...,
    ) -> bool: ...
    def create(
        self, prog: Any | str = ..., format: str = ..., encoding: Any | str = ...
    ) -> bytes: ...



# Stubs for networkx.algorithms.link_analysis.hits_alg (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def hits(
    G,
    max_iter: int = ...,
    tol: float = ...,
    nstart: Optional[Any] = ...,
    normalized: bool = ...,
): ...
def authority_matrix(G, nodelist: Optional[Any] = ...): ...
def hub_matrix(G, nodelist: Optional[Any] = ...): ...
def hits_numpy(G, normalized: bool = ...): ...
def hits_scipy(G, max_iter: int = ..., tol: float = ..., normalized: bool = ...): ...



# Stubs for networkx.algorithms.distance_measures (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def extrema_bounding(G, compute: str = ...): ...
def eccentricity(G, v: Optional[Any] = ..., sp: Optional[Any] = ...): ...
def diameter(G, e: Optional[Any] = ..., usebounds: bool = ...): ...
def periphery(G, e: Optional[Any] = ..., usebounds: bool = ...): ...
def radius(G, e: Optional[Any] = ..., usebounds: bool = ...): ...
def center(G, e: Optional[Any] = ..., usebounds: bool = ...): ...



# Stubs for networkx.classes.multidigraph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.classes.digraph import DiGraph
from networkx.classes.multigraph import MultiGraph

class MultiDiGraph(MultiGraph, DiGraph):
    edge_key_dict_factory: Any = ...
    def __init__(self, incoming_graph_data: Optional[Any] = ..., **attr) -> None: ...
    @property
    def adj(self): ...
    @property
    def succ(self): ...
    @property
    def pred(self): ...
    def add_edge(self, u_for_edge, v_for_edge, key: Optional[Any] = ..., **attr): ...
    def remove_edge(self, u, v, key: Optional[Any] = ...): ...
    @property
    def edges(self): ...
    out_edges: Any = ...
    @property
    def in_edges(self): ...
    @property
    def degree(self): ...
    @property
    def in_degree(self): ...
    @property
    def out_degree(self): ...
    def is_multigraph(self): ...
    def is_directed(self): ...
    def fresh_copy(self): ...
    def copy(self, as_view: bool = ...): ...
    def to_undirected(self, reciprocal: bool = ..., as_view: bool = ...): ...
    def subgraph(self, nodes): ...
    def reverse(self, copy: bool = ...): ...



# Stubs for networkx.algorithms.shortest_paths.generic (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def has_path(G, source, target): ...
def shortest_path(
    G,
    source: Optional[Any] = ...,
    target: Optional[Any] = ...,
    weight: Optional[Any] = ...,
): ...
def shortest_path_length(
    G,
    source: Optional[Any] = ...,
    target: Optional[Any] = ...,
    weight: Optional[Any] = ...,
): ...
def average_shortest_path_length(G, weight: Optional[Any] = ...): ...
def all_shortest_paths(G, source, target, weight: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.bipartite.basic (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def color(G): ...
def is_bipartite(G): ...
def is_bipartite_node_set(G, nodes): ...
def sets(G, top_nodes: Optional[Any] = ...): ...
def density(B, nodes): ...
def degrees(B, nodes, weight: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.graphical (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def is_graphical(sequence, method: str = ...): ...
def is_valid_degree_sequence_havel_hakimi(deg_sequence): ...
def is_valid_degree_sequence_erdos_gallai(deg_sequence): ...
def is_multigraphical(sequence): ...
def is_pseudographical(sequence): ...
def is_digraphical(in_sequence, out_sequence): ...



import os
import sys
from datetime import datetime
from pathlib import Path
from typing import List

import pytest
from pytest_httpserver import HTTPServer

from cwltool.pathmapper import PathMapper
from cwltool.utils import CWLObjectType


def test_http_path_mapping(tmp_path: Path) -> None:
    input_file_path = (
        "https://raw.githubusercontent.com/common-workflow-language/cwltool/main/tests/2.fasta"
    )
    base_file: List[CWLObjectType] = [
        {
            "class": "File",
            "location": "https://raw.githubusercontent.com/common-workflow-language/"
            "cwltool/main/tests/2.fasta",
            "basename": "chr20.fa",
        }
    ]
    pathmap = PathMapper(base_file, os.getcwd(), str(tmp_path))._pathmap

    assert input_file_path in pathmap
    assert os.path.exists(pathmap[input_file_path].resolved)

    with open(pathmap[input_file_path].resolved) as file:
        contents = file.read()

    assert ">Sequence 561 BP; 135 A; 106 C; 98 G; 222 T; 0 other;" in contents


@pytest.mark.skipif(sys.version_info < (3, 7), reason="timesout on CI")
def test_modification_date(tmp_path: Path) -> None:
    """Local copies of remote files should preserve last modification date."""
    # Initialize the server
    headers = {
        "Server": "nginx",
        "Date": "Mon, 27 Jun 2022 14:26:17 GMT",
        "Content-Type": "application/zip",
        "Content-Length": "123906",
        "Connection": "keep-alive",
        "Last-Modified": "Tue, 14 Dec 2021 14:23:30 GMT",
        "ETag": '"1e402-5d31beef49671"',
        "Accept-Ranges": "bytes",
        "Strict-Transport-Security": "max-age=31536000; includeSubDomains",
    }

    remote_file_name = "testfile.txt"

    with HTTPServer() as httpserver:
        httpserver.expect_request(f"/{remote_file_name}").respond_with_data(
            response_data="Hello World", headers=headers
        )
        location = httpserver.url_for(f"/{remote_file_name}")

        base_file: List[CWLObjectType] = [
            {
                "class": "File",
                "location": location,
                "basename": remote_file_name,
            }
        ]

        date_now = datetime.now()

        pathmap = PathMapper(base_file, os.getcwd(), str(tmp_path))._pathmap

        assert location in pathmap
        assert os.path.exists(pathmap[location].resolved)

        last_modified = os.path.getmtime(pathmap[location].resolved)

        assert date_now.timestamp() > last_modified



# Stubs for networkx.algorithms.centrality.eigenvector (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def eigenvector_centrality(
    G,
    max_iter: int = ...,
    tol: float = ...,
    nstart: Optional[Any] = ...,
    weight: Optional[Any] = ...,
): ...
def eigenvector_centrality_numpy(
    G, weight: Optional[Any] = ..., max_iter: int = ..., tol: int = ...
): ...



# Stubs for networkx.classes.reportviews (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from collections import Mapping, Set
from typing import Any, Optional

class NodeView(Mapping, Set):
    def __init__(self, graph) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __getitem__(self, n): ...
    def __contains__(self, n): ...
    def __call__(self, data: bool = ..., default: Optional[Any] = ...): ...
    def data(self, data: bool = ..., default: Optional[Any] = ...): ...

class NodeDataView(Set):
    def __init__(self, nodedict, data: bool = ..., default: Optional[Any] = ...) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, n): ...
    def __getitem__(self, n): ...

class DiDegreeView:
    def __init__(self, G, nbunch: Optional[Any] = ..., weight: Optional[Any] = ...) -> None: ...
    def __call__(self, nbunch: Optional[Any] = ..., weight: Optional[Any] = ...): ...
    def __getitem__(self, n): ...
    def __iter__(self): ...
    def __len__(self): ...

class DegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class OutDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class InDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class MultiDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class DiMultiDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class InMultiDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class OutMultiDegreeView(DiDegreeView):
    def __getitem__(self, n): ...
    def __iter__(self): ...

class OutEdgeDataView:
    def __init__(
        self,
        viewer,
        nbunch: Optional[Any] = ...,
        data: bool = ...,
        default: Optional[Any] = ...,
    ) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...

class EdgeDataView(OutEdgeDataView):
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...

class InEdgeDataView(OutEdgeDataView):
    def __iter__(self): ...
    def __contains__(self, e): ...

class OutMultiEdgeDataView(OutEdgeDataView):
    keys: Any = ...
    def __init__(
        self,
        viewer,
        nbunch: Optional[Any] = ...,
        data: bool = ...,
        keys: bool = ...,
        default: Optional[Any] = ...,
    ) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...

class MultiEdgeDataView(OutMultiEdgeDataView):
    def __iter__(self): ...
    def __contains__(self, e): ...

class InMultiEdgeDataView(OutMultiEdgeDataView):
    def __iter__(self): ...
    def __contains__(self, e): ...

class OutEdgeView(Set, Mapping):
    dataview: Any = ...
    def __init__(self, G) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...
    def __getitem__(self, e): ...
    def __call__(
        self,
        nbunch: Optional[Any] = ...,
        data: bool = ...,
        default: Optional[Any] = ...,
    ): ...
    def data(
        self,
        data: bool = ...,
        default: Optional[Any] = ...,
        nbunch: Optional[Any] = ...,
    ): ...

class EdgeView(OutEdgeView):
    dataview: Any = ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...

class InEdgeView(OutEdgeView):
    dataview: Any = ...
    def __init__(self, G) -> None: ...
    def __iter__(self): ...
    def __contains__(self, e): ...
    def __getitem__(self, e): ...

class OutMultiEdgeView(OutEdgeView):
    dataview: Any = ...
    def __len__(self): ...
    def __iter__(self): ...
    def __contains__(self, e): ...
    def __getitem__(self, e): ...
    def __call__(
        self,
        nbunch: Optional[Any] = ...,
        data: bool = ...,
        keys: bool = ...,
        default: Optional[Any] = ...,
    ): ...
    def data(
        self,
        data: bool = ...,
        keys: bool = ...,
        default: Optional[Any] = ...,
        nbunch: Optional[Any] = ...,
    ): ...

class MultiEdgeView(OutMultiEdgeView):
    dataview: Any = ...
    def __len__(self): ...
    def __iter__(self): ...

class InMultiEdgeView(OutMultiEdgeView):
    dataview: Any = ...
    def __init__(self, G) -> None: ...
    def __iter__(self): ...
    def __contains__(self, e): ...
    def __getitem__(self, e): ...



# Stubs for networkx.algorithms.tree.branchings (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from .recognition import *

def branching_weight(G, attr: str = ..., default: int = ...): ...
def greedy_branching(G, attr: str = ..., default: int = ..., kind: str = ...): ...

class MultiDiGraph_EdgeKey(networkx.MultiDiGraph):
    edge_index: Any = ...
    def __init__(self, incoming_graph_data: Optional[Any] = ..., **attr) -> None: ...
    def remove_node(self, n): ...
    def remove_nodes_from(self, nbunch): ...
    def fresh_copy(self): ...
    def add_edge(self, u_for_edge, v_for_edge, key_for_edge, **attr): ...
    def add_edges_from(self, ebunch_to_add, **attr): ...
    def remove_edge_with_key(self, key): ...
    def remove_edges_from(self, ebunch): ...

class Edmonds:
    G_original: Any = ...
    store: bool = ...
    edges: Any = ...
    template: Any = ...
    def __init__(self, G, seed: Optional[Any] = ...) -> None: ...
    def find_optimum(
        self, attr: str = ..., default: int = ..., kind: str = ..., style: str = ...
    ): ...

def maximum_branching(G, attr: str = ..., default: int = ...): ...
def minimum_branching(G, attr: str = ..., default: int = ...): ...
def maximum_spanning_arborescence(G, attr: str = ..., default: int = ...): ...
def minimum_spanning_arborescence(G, attr: str = ..., default: int = ...): ...



# Stubs for networkx.generators.directed (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def gn_graph(
    n,
    kernel: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    seed: Optional[Any] = ...,
): ...
def gnr_graph(n, p, create_using: Optional[Any] = ..., seed: Optional[Any] = ...): ...
def gnc_graph(n, create_using: Optional[Any] = ..., seed: Optional[Any] = ...): ...
def scale_free_graph(
    n,
    alpha: float = ...,
    beta: float = ...,
    gamma: float = ...,
    delta_in: float = ...,
    delta_out: int = ...,
    create_using: Optional[Any] = ...,
    seed: Optional[Any] = ...,
): ...
def random_k_out_graph(n, k, alpha, self_loops: bool = ..., seed: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.link_analysis (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.link_analysis.hits_alg import *
from networkx.algorithms.link_analysis.pagerank_alg import *



# Stubs for networkx.readwrite.json_graph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.readwrite.json_graph.adjacency import *
from networkx.readwrite.json_graph.cytoscape import *
from networkx.readwrite.json_graph.jit import *
from networkx.readwrite.json_graph.node_link import *
from networkx.readwrite.json_graph.tree import *



# Stubs for networkx.utils.misc (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

basestring = str
unicode = str

def is_string_like(obj): ...
def iterable(obj): ...
def flatten(obj, result: Optional[Any] = ...): ...
def is_list_of_ints(intlist): ...

PY2: Any

def make_str(x): ...
def generate_unique_node(): ...
def default_opener(filename): ...
def dict_to_numpy_array(d, mapping: Optional[Any] = ...): ...
def dict_to_numpy_array2(d, mapping: Optional[Any] = ...): ...
def dict_to_numpy_array1(d, mapping: Optional[Any] = ...): ...
def is_iterator(obj): ...
def arbitrary_element(iterable): ...
def consume(iterator): ...
def pairwise(iterable, cyclic: bool = ...): ...
def groups(many_to_one): ...
def to_tuple(x): ...
def create_random_state(random_state: Optional[Any] = ...): ...



# Stubs for networkx.linalg.graphmatrix (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def incidence_matrix(
    G,
    nodelist: Optional[Any] = ...,
    edgelist: Optional[Any] = ...,
    oriented: bool = ...,
    weight: Optional[Any] = ...,
): ...
def adjacency_matrix(G, nodelist: Optional[Any] = ..., weight: str = ...): ...

adj_matrix = adjacency_matrix



"""Tests of satisfying SoftwareRequirement via dependencies."""
import os
from pathlib import Path
from shutil import which
from types import ModuleType
from typing import Optional

import pytest

from cwltool.context import LoadingContext
from cwltool.load_tool import load_tool
from cwltool.software_requirements import get_container_from_software_requirements

from .util import get_data, get_main_output, get_tool_env, needs_docker

deps: Optional[ModuleType] = None
try:
    from galaxy.tool_util import deps  # type: ignore[no-redef]
except ImportError:
    pass


@needs_docker
@pytest.mark.skipif(not deps, reason="galaxy-tool-util is not installed")
def test_biocontainers(tmp_path: Path) -> None:
    wflow = get_data("tests/seqtk_seq.cwl")
    job = get_data("tests/seqtk_seq_job.json")
    error_code, _, _ = get_main_output(
        ["--outdir", str(tmp_path), "--beta-use-biocontainers", wflow, job]
    )

    assert error_code == 0


@needs_docker
@pytest.mark.skipif(not deps, reason="galaxy-tool-util is not installed")
def test_biocontainers_resolution(tmp_path: Path) -> None:
    """Confirm expected container name for --beta-use-biocontainers."""
    tool = load_tool(get_data("tests/seqtk_seq.cwl"), LoadingContext())
    assert (
        get_container_from_software_requirements(True, tool) == "quay.io/biocontainers/seqtk:r93--0"
    )


@pytest.mark.skipif(not deps, reason="galaxy-tool-util is not installed")
def test_bioconda(tmp_path: Path) -> None:
    wflow = get_data("tests/seqtk_seq.cwl")
    job = get_data("tests/seqtk_seq_job.json")
    error_code, _, stderr = get_main_output(
        ["--outdir", str(tmp_path), "--beta-conda-dependencies", "--debug", wflow, job]
    )

    assert error_code == 0, stderr


@pytest.mark.skipif(not deps, reason="galaxy-tool-util is not installed")
@pytest.mark.skipif(not which("modulecmd"), reason="modulecmd not installed")
def test_modules(monkeypatch: pytest.MonkeyPatch) -> None:
    """Do a basic smoke test using environment modules to satisfy a SoftwareRequirement."""
    wflow = get_data("tests/random_lines.cwl")
    job = get_data("tests/random_lines_job.json")
    monkeypatch.setenv("MODULEPATH", os.path.join(os.getcwd(), "tests/test_deps_env/modulefiles"))
    error_code, _, stderr = get_main_output(
        [
            "--beta-dependency-resolvers-configuration",
            "tests/test_deps_env_modules_resolvers_conf.yml",
            "--debug",
            wflow,
            job,
        ]
    )

    assert error_code == 0, stderr


@pytest.mark.skipif(not deps, reason="galaxy-tool-util is not installed")
@pytest.mark.skipif(not which("modulecmd"), reason="modulecmd not installed")
def test_modules_environment(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Check that the environment variables set by a module are being propagated correctly.

    Do so by  by running `env` as the tool and parsing its output.
    """
    monkeypatch.setenv("MODULEPATH", os.path.join(os.getcwd(), "tests/test_deps_env/modulefiles"))
    tool_env = get_tool_env(
        tmp_path,
        [
            "--beta-dependency-resolvers-configuration",
            get_data("tests/test_deps_env_modules_resolvers_conf.yml"),
        ],
        get_data("tests/env_with_software_req.yml"),
    )

    assert tool_env["TEST_VAR_MODULE"] == "environment variable ends in space "
    tool_path = tool_env["PATH"].split(":")
    assert get_data("tests/test_deps_env/random-lines/1.0/scripts") in tool_path



# Stubs for networkx.algorithms.richclub (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def rich_club_coefficient(G, normalized: bool = ..., Q: int = ...): ...



# Stubs for networkx.algorithms.bipartite.spectral (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def spectral_bipartivity(G, nodes: Optional[Any] = ..., weight: str = ...): ...



import json
from pathlib import Path

from cwltool.context import RuntimeContext
from cwltool.executors import MultithreadedJobExecutor
from cwltool.factory import Factory

from .util import get_data, needs_docker


@needs_docker
def test_sequential_workflow(tmp_path: Path) -> None:
    test_file = "tests/wf/count-lines1-wf.cwl"
    executor = MultithreadedJobExecutor()
    runtime_context = RuntimeContext()
    runtime_context.outdir = str(tmp_path)
    runtime_context.select_resources = executor.select_resources
    factory = Factory(executor, None, runtime_context)
    echo = factory.make(get_data(test_file))
    file_contents = {"class": "File", "location": get_data("tests/wf/whale.txt")}
    assert echo(file1=file_contents) == {"count_output": 16}


@needs_docker
def test_scattered_workflow() -> None:
    test_file = "tests/wf/scatter-wf4.cwl"
    job_file = "tests/wf/scatter-job2.json"
    factory = Factory(MultithreadedJobExecutor())
    echo = factory.make(get_data(test_file))
    with open(get_data(job_file)) as job:
        assert echo(**json.load(job)) == {"out": ["foo one three", "foo two four"]}



# Stubs for networkx.algorithms.operators.unary (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def complement(G): ...
def reverse(G, copy: bool = ...): ...



# Stubs for networkx.algorithms.core (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def core_number(G): ...

find_cores = core_number

def k_core(G, k: Optional[Any] = ..., core_number: Optional[Any] = ...): ...
def k_shell(G, k: Optional[Any] = ..., core_number: Optional[Any] = ...): ...
def k_crust(G, k: Optional[Any] = ..., core_number: Optional[Any] = ...): ...
def k_corona(G, k, core_number: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.connectivity.edge_augmentation (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from collections import namedtuple
from typing import Any, Optional

def is_k_edge_connected(G, k): ...
def is_locally_k_edge_connected(G, s, t, k): ...
def k_edge_augmentation(
    G, k, avail: Optional[Any] = ..., weight: Optional[Any] = ..., partial: bool = ...
): ...

# MetaEdge = namedtuple('MetaEdge', <ERROR>)



# Stubs for networkx.readwrite.gpickle (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def write_gpickle(G, path, protocol: Any = ...): ...
def read_gpickle(path): ...



# Stubs for networkx.algorithms.traversal (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .beamsearch import *
from .breadth_first_search import *
from .depth_first_search import *
from .edgedfs import *



# Stubs for networkx.generators.social (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def karate_club_graph(): ...
def davis_southern_women_graph(): ...
def florentine_families_graph(): ...



"""Support for executing Docker format containers using Singularity {2,3}.x or Apptainer 1.x."""

import os
import os.path
from subprocess import DEVNULL, PIPE, Popen, TimeoutExpired  # nosec
from typing import Optional

_USERNS: Optional[bool] = None


def singularity_supports_userns() -> bool:
    """Confirm if the version of Singularity install supports the --userns flag."""
    global _USERNS  # pylint: disable=global-statement
    if _USERNS is None:
        try:
            hello_image = os.path.join(os.path.dirname(__file__), "hello.simg")
            result = Popen(  # nosec
                ["singularity", "exec", "--userns", hello_image, "true"],
                stderr=PIPE,
                stdout=DEVNULL,
                universal_newlines=True,
            ).communicate(timeout=60)[1]
            _USERNS = (
                "No valid /bin/sh" in result
                or "/bin/sh doesn't exist in container" in result
                or "executable file not found in" in result
            )
        except TimeoutExpired:
            _USERNS = False
    return _USERNS



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class TIME(DefinedNamespace):
    generalDay: URIRef
    generalMonth: URIRef
    generalYear: URIRef
    DateTimeDescription: URIRef
    DateTimeInterval: URIRef
    DayOfWeek: URIRef
    Duration: URIRef
    DurationDescription: URIRef
    GeneralDateTimeDescription: URIRef
    GeneralDurationDescription: URIRef
    Instant: URIRef
    Interval: URIRef
    MonthOfYear: URIRef
    ProperInterval: URIRef
    TRS: URIRef
    TemporalDuration: URIRef
    TemporalEntity: URIRef
    TemporalPosition: URIRef
    TemporalUnit: URIRef
    TimePosition: URIRef
    TimeZone: URIRef
    day: URIRef
    dayOfYear: URIRef
    days: URIRef
    hasXSDDuration: URIRef
    hour: URIRef
    hours: URIRef
    inXSDDate: URIRef
    inXSDDateTimeStamp: URIRef
    inXSDgYear: URIRef
    inXSDgYearMonth: URIRef
    minute: URIRef
    minutes: URIRef
    month: URIRef
    months: URIRef
    nominalPosition: URIRef
    numericDuration: URIRef
    numericPosition: URIRef
    second: URIRef
    seconds: URIRef
    week: URIRef
    weeks: URIRef
    year: URIRef
    years: URIRef
    January: URIRef
    Year: URIRef
    inXSDDateTime: URIRef
    xsdDateTime: URIRef
    hasTRS: URIRef
    after: URIRef
    dayOfWeek: URIRef
    hasBeginning: URIRef
    hasDateTimeDescription: URIRef
    hasDuration: URIRef
    hasDurationDescription: URIRef
    hasEnd: URIRef
    hasTemporalDuration: URIRef
    hasTime: URIRef
    inDateTime: URIRef
    inTemporalPosition: URIRef
    inTimePosition: URIRef
    inside: URIRef
    intervalAfter: URIRef
    intervalBefore: URIRef
    intervalContains: URIRef
    intervalDisjoint: URIRef
    intervalDuring: URIRef
    intervalEquals: URIRef
    intervalFinishedBy: URIRef
    intervalFinishes: URIRef
    intervalIn: URIRef
    intervalMeets: URIRef
    intervalMetBy: URIRef
    intervalOverlappedBy: URIRef
    intervalOverlaps: URIRef
    intervalStartedBy: URIRef
    intervalStarts: URIRef
    monthOfYear: URIRef
    timeZone: URIRef
    unitType: URIRef
    before: URIRef
    Friday: URIRef
    Monday: URIRef
    Saturday: URIRef
    Sunday: URIRef
    Thursday: URIRef
    Tuesday: URIRef
    Wednesday: URIRef
    unitDay: URIRef
    unitHour: URIRef
    unitMinute: URIRef
    unitMonth: URIRef
    unitSecond: URIRef
    unitWeek: URIRef
    unitYear: URIRef



# Stubs for networkx.algorithms.bipartite.matrix (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def biadjacency_matrix(
    G,
    row_order,
    column_order: Optional[Any] = ...,
    dtype: Optional[Any] = ...,
    weight: str = ...,
    format: str = ...,
): ...
def from_biadjacency_matrix(A, create_using: Optional[Any] = ..., edge_attribute: str = ...): ...



import subprocess
import time
import pkg_resources
from setuptools.command.egg_info import egg_info

SETUPTOOLS_VER = pkg_resources.get_distribution(
    "setuptools").version.split('.')

RECENT_SETUPTOOLS = int(SETUPTOOLS_VER[0]) > 40 or (
    int(SETUPTOOLS_VER[0]) == 40 and int(SETUPTOOLS_VER[1]) > 0) or (
        int(SETUPTOOLS_VER[0]) == 40 and int(SETUPTOOLS_VER[1]) == 0 and
        int(SETUPTOOLS_VER[2]) > 0)

class EggInfoFromGit(egg_info):
    """Tag the build with git commit timestamp.

    If a build tag has already been set (e.g., "egg_info -b", building
    from source package), leave it alone.
    """

    def git_timestamp_tag(self):
        gitinfo = subprocess.check_output(
            ['git', 'log', '--first-parent', '--max-count=1',
             '--format=format:%ct', '.']).strip()
        return time.strftime('.%Y%m%d%H%M%S', time.gmtime(int(gitinfo)))

    def tags(self):
        if self.tag_build is None:
            try:
                self.tag_build = self.git_timestamp_tag()
            except subprocess.CalledProcessError:
                pass
        return egg_info.tags(self)

    if RECENT_SETUPTOOLS:
        vtags = property(tags)



# Stubs for networkx.algorithms.community.community_generators (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def LFR_benchmark_graph(
    n,
    tau1,
    tau2,
    mu,
    average_degree: Optional[Any] = ...,
    min_degree: Optional[Any] = ...,
    max_degree: Optional[Any] = ...,
    min_community: Optional[Any] = ...,
    max_community: Optional[Any] = ...,
    tol: float = ...,
    max_iters: int = ...,
    seed: Optional[Any] = ...,
): ...



# Stubs for networkx.readwrite (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.readwrite.adjlist import *
from networkx.readwrite.edgelist import *
from networkx.readwrite.gexf import *
from networkx.readwrite.gml import *
from networkx.readwrite.gpickle import *
from networkx.readwrite.graph6 import *
from networkx.readwrite.graphml import *
from networkx.readwrite.json_graph import *
from networkx.readwrite.leda import *
from networkx.readwrite.multiline_adjlist import *
from networkx.readwrite.nx_shp import *
from networkx.readwrite.nx_yaml import *
from networkx.readwrite.pajek import *
from networkx.readwrite.sparse6 import *



# Stubs for networkx.classes.function (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def nodes(G): ...
def edges(G, nbunch: Optional[Any] = ...): ...
def degree(G, nbunch: Optional[Any] = ..., weight: Optional[Any] = ...): ...
def neighbors(G, n): ...
def number_of_nodes(G): ...
def number_of_edges(G): ...
def density(G): ...
def degree_histogram(G): ...
def is_directed(G): ...
def freeze(G): ...
def is_frozen(G): ...
def add_star(G_to_add_to, nodes_for_star, **attr): ...
def add_path(G_to_add_to, nodes_for_path, **attr): ...
def add_cycle(G_to_add_to, nodes_for_cycle, **attr): ...
def subgraph(G, nbunch): ...
def induced_subgraph(G, nbunch): ...
def edge_subgraph(G, edges): ...
def restricted_view(G, nodes, edges): ...
def reverse_view(digraph): ...
def to_directed(graph): ...
def to_undirected(graph): ...
def create_empty_copy(G, with_data: bool = ...): ...
def info(G, n: Optional[Any] = ...): ...
def set_node_attributes(G, values, name: Optional[Any] = ...): ...
def get_node_attributes(G, name): ...
def set_edge_attributes(G, values, name: Optional[Any] = ...): ...
def get_edge_attributes(G, name): ...
def all_neighbors(graph, node): ...
def non_neighbors(graph, node): ...
def non_edges(graph): ...
def common_neighbors(G, u, v): ...
def is_weighted(G, edge: Optional[Any] = ..., weight: str = ...): ...
def is_negatively_weighted(G, edge: Optional[Any] = ..., weight: str = ...): ...
def is_empty(G): ...
def nodes_with_selfloops(G): ...
def selfloop_edges(G, data: bool = ..., keys: bool = ..., default: Optional[Any] = ...): ...
def number_of_selfloops(G): ...



# Stubs for networkx.release (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

basedir: Any

def write_versionfile(): ...
def get_revision(): ...
def get_info(dynamic: bool = ...): ...

name: str
major: str
minor: str
dev: bool
description: str
long_description: str
license: str
authors: Any
maintainer: str
maintainer_email: str
url: str
download_url: str
platforms: Any
keywords: Any
classifiers: Any
date: Any
date_info: Any
version: Any
version_info: Any
vcs_info: Any



# Stubs for networkx.readwrite.json_graph.cytoscape (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def cytoscape_data(G, attrs: Optional[Any] = ...): ...
def cytoscape_graph(data, attrs: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.connectivity.kcomponents (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.algorithms.flow import edmonds_karp

default_flow_func = edmonds_karp

def k_components(G, flow_func: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.dominance (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def immediate_dominators(G, start): ...
def dominance_frontiers(G, start): ...



# Stubs for networkx.algorithms.boundary (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def edge_boundary(
    G,
    nbunch1,
    nbunch2: Optional[Any] = ...,
    data: bool = ...,
    keys: bool = ...,
    default: Optional[Any] = ...,
): ...
def node_boundary(G, nbunch1, nbunch2: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.isomorphism.isomorph (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def could_be_isomorphic(G1, G2): ...

graph_could_be_isomorphic = could_be_isomorphic

def fast_could_be_isomorphic(G1, G2): ...

fast_graph_could_be_isomorphic = fast_could_be_isomorphic

def faster_could_be_isomorphic(G1, G2): ...

faster_graph_could_be_isomorphic = faster_could_be_isomorphic

def is_isomorphic(G1, G2, node_match: Optional[Any] = ..., edge_match: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.centrality.degree_alg (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def degree_centrality(G): ...
def in_degree_centrality(G): ...
def out_degree_centrality(G): ...



# Stubs for networkx.utils.random_sequence (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def powerlaw_sequence(n, exponent: float = ...): ...
def zipf_rv(alpha, xmin: int = ..., seed: Optional[Any] = ...): ...
def cumulative_distribution(distribution): ...
def discrete_sequence(n, distribution: Optional[Any] = ..., cdistribution: Optional[Any] = ...): ...
def random_weighted_sample(mapping, k): ...
def weighted_choice(mapping): ...



# Stubs for networkx.drawing.nx_pylab (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def draw(G, pos: Optional[Any] = ..., ax: Optional[Any] = ..., **kwds): ...
def draw_networkx(
    G, pos: Optional[Any] = ..., arrows: bool = ..., with_labels: bool = ..., **kwds
): ...
def draw_networkx_nodes(
    G,
    pos,
    nodelist: Optional[Any] = ...,
    node_size: int = ...,
    node_color: str = ...,
    node_shape: str = ...,
    alpha: float = ...,
    cmap: Optional[Any] = ...,
    vmin: Optional[Any] = ...,
    vmax: Optional[Any] = ...,
    ax: Optional[Any] = ...,
    linewidths: Optional[Any] = ...,
    edgecolors: Optional[Any] = ...,
    label: Optional[Any] = ...,
    **kwds,
): ...
def draw_networkx_edges(
    G,
    pos,
    edgelist: Optional[Any] = ...,
    width: float = ...,
    edge_color: str = ...,
    style: str = ...,
    alpha: float = ...,
    arrowstyle: str = ...,
    arrowsize: int = ...,
    edge_cmap: Optional[Any] = ...,
    edge_vmin: Optional[Any] = ...,
    edge_vmax: Optional[Any] = ...,
    ax: Optional[Any] = ...,
    arrows: bool = ...,
    label: Optional[Any] = ...,
    node_size: int = ...,
    nodelist: Optional[Any] = ...,
    node_shape: str = ...,
    **kwds,
): ...
def draw_networkx_labels(
    G,
    pos,
    labels: Optional[Any] = ...,
    font_size: int = ...,
    font_color: str = ...,
    font_family: str = ...,
    font_weight: str = ...,
    alpha: float = ...,
    bbox: Optional[Any] = ...,
    ax: Optional[Any] = ...,
    **kwds,
): ...
def draw_networkx_edge_labels(
    G,
    pos,
    edge_labels: Optional[Any] = ...,
    label_pos: float = ...,
    font_size: int = ...,
    font_color: str = ...,
    font_family: str = ...,
    font_weight: str = ...,
    alpha: float = ...,
    bbox: Optional[Any] = ...,
    ax: Optional[Any] = ...,
    rotate: bool = ...,
    **kwds,
): ...
def draw_circular(G, **kwargs): ...
def draw_kamada_kawai(G, **kwargs): ...
def draw_random(G, **kwargs): ...
def draw_spectral(G, **kwargs): ...
def draw_spring(G, **kwargs): ...
def draw_shell(G, **kwargs): ...



# Stubs for networkx.algorithms.community.quality (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx import NetworkXError

class NotAPartition(NetworkXError):
    def __init__(self, G, collection) -> None: ...

def performance(G, partition): ...
def coverage(G, partition): ...
def modularity(G, communities, weight: str = ...): ...



# Stubs for networkx.algorithms.minors (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def quotient_graph(
    G,
    partition,
    edge_relation: Optional[Any] = ...,
    node_data: Optional[Any] = ...,
    edge_data: Optional[Any] = ...,
    relabel: bool = ...,
    create_using: Optional[Any] = ...,
): ...
def contracted_nodes(G, u, v, self_loops: bool = ...): ...

identified_nodes = contracted_nodes

def contracted_edge(G, edge, self_loops: bool = ...): ...



# Stubs for networkx.readwrite.multiline_adjlist (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def generate_multiline_adjlist(G, delimiter: str = ...): ...
def write_multiline_adjlist(
    G, path, delimiter: str = ..., comments: str = ..., encoding: str = ...
): ...
def parse_multiline_adjlist(
    lines,
    comments: str = ...,
    delimiter: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    nodetype: Optional[Any] = ...,
    edgetype: Optional[Any] = ...,
): ...
def read_multiline_adjlist(
    path,
    comments: str = ...,
    delimiter: Optional[Any] = ...,
    create_using: Optional[Any] = ...,
    nodetype: Optional[Any] = ...,
    edgetype: Optional[Any] = ...,
    encoding: str = ...,
): ...



# Stubs for graphviz._compat (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

PY2: Any
string_classes: Any
text_type = unicode

def iteritems(d): ...
def makedirs(name, mode: int = ..., exist_ok: bool = ...): ...
def stderr_write_binary(data): ...

text_type = str



from _typeshed import Incomplete
from requests import Session

def CacheControl(
    sess: Session,
    cache: Incomplete | None = ...,
    cache_etags: bool = ...,
    serializer: Incomplete | None = ...,
    heuristic: Incomplete | None = ...,
    controller_class: Incomplete | None = ...,
    adapter_class: Incomplete | None = ...,
    cacheable_methods: Incomplete | None = ...,
) -> Session: ...



# Stubs for networkx.generators.classic (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def full_rary_tree(r, n, create_using: Optional[Any] = ...): ...
def balanced_tree(r, h, create_using: Optional[Any] = ...): ...
def barbell_graph(m1, m2, create_using: Optional[Any] = ...): ...
def complete_graph(n, create_using: Optional[Any] = ...): ...
def circular_ladder_graph(n, create_using: Optional[Any] = ...): ...
def circulant_graph(n, offsets, create_using: Optional[Any] = ...): ...
def cycle_graph(n, create_using: Optional[Any] = ...): ...
def dorogovtsev_goltsev_mendes_graph(n, create_using: Optional[Any] = ...): ...
def empty_graph(n: int = ..., create_using: Optional[Any] = ...): ...
def ladder_graph(n, create_using: Optional[Any] = ...): ...
def lollipop_graph(m, n, create_using: Optional[Any] = ...): ...
def null_graph(create_using: Optional[Any] = ...): ...
def path_graph(n, create_using: Optional[Any] = ...): ...
def star_graph(n, create_using: Optional[Any] = ...): ...
def trivial_graph(create_using: Optional[Any] = ...): ...
def turan_graph(n, r): ...
def wheel_graph(n, create_using: Optional[Any] = ...): ...
def complete_multipartite_graph(*subset_sizes): ...



# Stubs for networkx.generators.mycielski (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def mycielskian(G, iterations: int = ...): ...
def mycielski_graph(n): ...



# Stubs for networkx.algorithms.swap (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def double_edge_swap(G, nswap: int = ..., max_tries: int = ...): ...
def connected_double_edge_swap(G, nswap: int = ..., _window_threshold: int = ...): ...



"""Classes and methods relevant for all CWL Process types."""
import abc
import copy
import functools
import hashlib
import json
import logging
import math
import os
import shutil
import stat
import textwrap
import urllib.parse
import uuid
from os import scandir
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Sized,
    Tuple,
    Type,
    Union,
    cast,
)

from cwl_utils import expression
from mypy_extensions import mypyc_attr
from pkg_resources import resource_stream
from rdflib import Graph
from schema_salad.avro.schema import (
    Names,
    Schema,
    SchemaParseException,
    make_avsc_object,
)
from schema_salad.exceptions import ValidationException
from schema_salad.ref_resolver import Loader, file_uri, uri_file_path
from schema_salad.schema import load_schema, make_avro_schema, make_valid_avro
from schema_salad.sourceline import SourceLine, strip_dup_lineno
from schema_salad.utils import convert_to_dict
from schema_salad.validate import avro_type_name, validate_ex

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from .builder import INPUT_OBJ_VOCAB, Builder
from .context import LoadingContext, RuntimeContext, getdefault
from .errors import UnsupportedRequirement, WorkflowException
from .loghandler import _logger
from .mpi import MPIRequirementName
from .pathmapper import MapperEnt, PathMapper
from .secrets import SecretStore
from .stdfsaccess import StdFsAccess
from .update import INTERNAL_VERSION, ORIGINAL_CWLVERSION
from .utils import (
    CWLObjectType,
    CWLOutputAtomType,
    CWLOutputType,
    HasReqsHints,
    JobsGeneratorType,
    LoadListingType,
    OutputCallbackType,
    adjustDirObjs,
    aslist,
    cmp_like_py2,
    ensure_writable,
    get_listing,
    normalizeFilesDirs,
    random_outdir,
    visit_class,
)
from .validate_js import validate_js_expressions

if TYPE_CHECKING:
    from .provenance_profile import ProvenanceProfile


class LogAsDebugFilter(logging.Filter):
    def __init__(self, name: str, parent: logging.Logger) -> None:
        """Initialize."""
        name = str(name)
        super().__init__(name)
        self.parent = parent

    def filter(self, record: logging.LogRecord) -> bool:
        return self.parent.isEnabledFor(logging.DEBUG)


_logger_validation_warnings = logging.getLogger("cwltool.validation_warnings")
_logger_validation_warnings.setLevel(_logger.getEffectiveLevel())
_logger_validation_warnings.addFilter(LogAsDebugFilter("cwltool.validation_warnings", _logger))

supportedProcessRequirements = [
    "DockerRequirement",
    "SchemaDefRequirement",
    "EnvVarRequirement",
    "ScatterFeatureRequirement",
    "SubworkflowFeatureRequirement",
    "MultipleInputFeatureRequirement",
    "InlineJavascriptRequirement",
    "ShellCommandRequirement",
    "StepInputExpressionRequirement",
    "ResourceRequirement",
    "InitialWorkDirRequirement",
    "ToolTimeLimit",
    "WorkReuse",
    "NetworkAccess",
    "InplaceUpdateRequirement",
    "LoadListingRequirement",
    MPIRequirementName,
    "http://commonwl.org/cwltool#TimeLimit",
    "http://commonwl.org/cwltool#WorkReuse",
    "http://commonwl.org/cwltool#NetworkAccess",
    "http://commonwl.org/cwltool#LoadListingRequirement",
    "http://commonwl.org/cwltool#InplaceUpdateRequirement",
    "http://commonwl.org/cwltool#CUDARequirement",
]

cwl_files = (
    "Workflow.yml",
    "CommandLineTool.yml",
    "CommonWorkflowLanguage.yml",
    "Process.yml",
    "Operation.yml",
    "concepts.md",
    "contrib.md",
    "intro.md",
    "invocation.md",
)

salad_files = (
    "metaschema.yml",
    "metaschema_base.yml",
    "salad.md",
    "field_name.yml",
    "import_include.md",
    "link_res.yml",
    "ident_res.yml",
    "vocab_res.yml",
    "vocab_res.yml",
    "field_name_schema.yml",
    "field_name_src.yml",
    "field_name_proc.yml",
    "ident_res_schema.yml",
    "ident_res_src.yml",
    "ident_res_proc.yml",
    "link_res_schema.yml",
    "link_res_src.yml",
    "link_res_proc.yml",
    "vocab_res_schema.yml",
    "vocab_res_src.yml",
    "vocab_res_proc.yml",
)

SCHEMA_CACHE: Dict[
    str, Tuple[Loader, Union[Names, SchemaParseException], CWLObjectType, Loader]
] = {}
SCHEMA_FILE: Optional[CWLObjectType] = None
SCHEMA_DIR: Optional[CWLObjectType] = None
SCHEMA_ANY: Optional[CWLObjectType] = None

custom_schemas: Dict[str, Tuple[str, str]] = {}


def use_standard_schema(version: str) -> None:
    if version in custom_schemas:
        del custom_schemas[version]
    if version in SCHEMA_CACHE:
        del SCHEMA_CACHE[version]


def use_custom_schema(version: str, name: str, text: str) -> None:
    custom_schemas[version] = (name, text)
    if version in SCHEMA_CACHE:
        del SCHEMA_CACHE[version]


def get_schema(
    version: str,
) -> Tuple[Loader, Union[Names, SchemaParseException], CWLObjectType, Loader]:
    if version in SCHEMA_CACHE:
        return SCHEMA_CACHE[version]

    cache: Dict[str, Union[str, Graph, bool]] = {}
    version = version.split("#")[-1]
    if ".dev" in version:
        version = ".".join(version.split(".")[:-1])
    for f in cwl_files:
        try:
            res = resource_stream(__name__, f"schemas/{version}/{f}")
            cache["https://w3id.org/cwl/" + f] = res.read().decode("UTF-8")
            res.close()
        except OSError:
            pass

    for f in salad_files:
        try:
            res = resource_stream(
                __name__,
                f"schemas/{version}/salad/schema_salad/metaschema/{f}",
            )
            cache["https://w3id.org/cwl/salad/schema_salad/metaschema/" + f] = res.read().decode(
                "UTF-8"
            )
            res.close()
        except OSError:
            pass

    if version in custom_schemas:
        cache[custom_schemas[version][0]] = custom_schemas[version][1]
        SCHEMA_CACHE[version] = load_schema(custom_schemas[version][0], cache=cache)
    else:
        SCHEMA_CACHE[version] = load_schema(
            "https://w3id.org/cwl/CommonWorkflowLanguage.yml", cache=cache
        )

    return SCHEMA_CACHE[version]


def shortname(inputid: str) -> str:
    d = urllib.parse.urlparse(inputid)
    if d.fragment:
        return d.fragment.split("/")[-1]
    return d.path.split("/")[-1]


def stage_files(
    pathmapper: PathMapper,
    stage_func: Optional[Callable[[str, str], None]] = None,
    ignore_writable: bool = False,
    symlink: bool = True,
    secret_store: Optional[SecretStore] = None,
    fix_conflicts: bool = False,
) -> None:
    """
    Link or copy files to their targets. Create them as needed.

    :raises WorkflowException: if there is a file staging conflict
    """
    items = pathmapper.items() if not symlink else pathmapper.items_exclude_children()
    targets: Dict[str, MapperEnt] = {}
    for key, entry in items:
        if "File" not in entry.type:
            continue
        if entry.target not in targets:
            targets[entry.target] = entry
        elif targets[entry.target].resolved != entry.resolved:
            if fix_conflicts:
                # find first key that does not clash with an existing entry in targets
                # start with entry.target + '_' + 2 and then keep incrementing
                # the number till there is no clash
                i = 2
                tgt = f"{entry.target}_{i}"
                while tgt in targets:
                    i += 1
                    tgt = f"{entry.target}_{i}"
                targets[tgt] = pathmapper.update(key, entry.resolved, tgt, entry.type, entry.staged)
            else:
                raise WorkflowException(
                    "File staging conflict, trying to stage both %s and %s to the same target %s"
                    % (targets[entry.target].resolved, entry.resolved, entry.target)
                )

    for key, entry in items:
        if not entry.staged:
            continue
        if not os.path.exists(os.path.dirname(entry.target)):
            os.makedirs(os.path.dirname(entry.target))
        if entry.type in ("File", "Directory") and os.path.exists(entry.resolved):
            if symlink:  # Use symlink func if allowed
                os.symlink(entry.resolved, entry.target)
            elif stage_func is not None:
                stage_func(entry.resolved, entry.target)
        elif (
            entry.type == "Directory"
            and not os.path.exists(entry.target)
            and entry.resolved.startswith("_:")
        ):
            os.makedirs(entry.target)
        elif entry.type == "WritableFile" and not ignore_writable:
            shutil.copy(entry.resolved, entry.target)
            ensure_writable(entry.target)
        elif entry.type == "WritableDirectory" and not ignore_writable:
            if entry.resolved.startswith("_:"):
                os.makedirs(entry.target)
            else:
                shutil.copytree(entry.resolved, entry.target)
                ensure_writable(entry.target, include_root=True)
        elif entry.type == "CreateFile" or entry.type == "CreateWritableFile":
            with open(entry.target, "w") as new:
                if secret_store is not None:
                    new.write(cast(str, secret_store.retrieve(entry.resolved)))
                else:
                    new.write(entry.resolved)
            if entry.type == "CreateFile":
                os.chmod(entry.target, stat.S_IRUSR)  # Read only
            else:  # it is a "CreateWritableFile"
                ensure_writable(entry.target)
            pathmapper.update(key, entry.target, entry.target, entry.type, entry.staged)


def relocateOutputs(
    outputObj: CWLObjectType,
    destination_path: str,
    source_directories: Set[str],
    action: str,
    fs_access: StdFsAccess,
    compute_checksum: bool = True,
    path_mapper: Type[PathMapper] = PathMapper,
) -> CWLObjectType:
    adjustDirObjs(outputObj, functools.partial(get_listing, fs_access, recursive=True))

    if action not in ("move", "copy"):
        return outputObj

    def _collectDirEntries(
        obj: Union[CWLObjectType, MutableSequence[CWLObjectType], None]
    ) -> Iterator[CWLObjectType]:
        if isinstance(obj, dict):
            if obj.get("class") in ("File", "Directory"):
                yield obj
            else:
                for sub_obj in obj.values():
                    yield from _collectDirEntries(sub_obj)
        elif isinstance(obj, MutableSequence):
            for sub_obj in obj:
                yield from _collectDirEntries(sub_obj)

    def _relocate(src: str, dst: str) -> None:
        src = fs_access.realpath(src)
        dst = fs_access.realpath(dst)

        if src == dst:
            return

        # If the source is not contained in source_directories we're not allowed to delete it
        src_can_deleted = any(os.path.commonprefix([p, src]) == p for p in source_directories)

        _action = "move" if action == "move" and src_can_deleted else "copy"

        if _action == "move":
            _logger.debug("Moving %s to %s", src, dst)
            if fs_access.isdir(src) and fs_access.isdir(dst):
                # merge directories
                for dir_entry in scandir(src):
                    _relocate(dir_entry.path, fs_access.join(dst, dir_entry.name))
            else:
                shutil.move(src, dst)

        elif _action == "copy":
            _logger.debug("Copying %s to %s", src, dst)
            if fs_access.isdir(src):
                if os.path.isdir(dst):
                    shutil.rmtree(dst)
                elif os.path.isfile(dst):
                    os.unlink(dst)
                shutil.copytree(src, dst)
            else:
                shutil.copy2(src, dst)

    def _realpath(
        ob: CWLObjectType,
    ) -> None:  # should be type Union[CWLFile, CWLDirectory]
        location = cast(str, ob["location"])
        if location.startswith("file:"):
            ob["location"] = file_uri(os.path.realpath(uri_file_path(location)))
        elif location.startswith("/"):
            ob["location"] = os.path.realpath(location)
        elif not location.startswith("_:") and ":" in location:
            ob["location"] = file_uri(fs_access.realpath(location))

    outfiles = list(_collectDirEntries(outputObj))
    visit_class(outfiles, ("File", "Directory"), _realpath)
    pm = path_mapper(outfiles, "", destination_path, separateDirs=False)
    stage_files(pm, stage_func=_relocate, symlink=False, fix_conflicts=True)

    def _check_adjust(a_file: CWLObjectType) -> CWLObjectType:
        a_file["location"] = file_uri(pm.mapper(cast(str, a_file["location"]))[1])
        if "contents" in a_file:
            del a_file["contents"]
        return a_file

    visit_class(outputObj, ("File", "Directory"), _check_adjust)

    if compute_checksum:
        visit_class(outputObj, ("File",), functools.partial(compute_checksums, fs_access))
    return outputObj


def cleanIntermediate(output_dirs: Iterable[str]) -> None:
    for a in output_dirs:
        if os.path.exists(a):
            _logger.debug("Removing intermediate output directory %s", a)
            shutil.rmtree(a, True)


def add_sizes(fsaccess: StdFsAccess, obj: CWLObjectType) -> None:
    if "location" in obj:
        try:
            if "size" not in obj:
                obj["size"] = fsaccess.size(cast(str, obj["location"]))
        except OSError:
            pass
    elif "contents" in obj:
        obj["size"] = len(cast(Sized, obj["contents"]))
    return  # best effort


def fill_in_defaults(
    inputs: List[CWLObjectType],
    job: CWLObjectType,
    fsaccess: StdFsAccess,
) -> None:
    """
    For each missing input in the input object, copy over the default.

    :raises WorkflowException: if a required input parameter is missing
    """
    debug = _logger.isEnabledFor(logging.DEBUG)
    for e, inp in enumerate(inputs):
        with SourceLine(inputs, e, WorkflowException, debug):
            fieldname = shortname(cast(str, inp["id"]))
            if job.get(fieldname) is not None:
                pass
            elif job.get(fieldname) is None and "default" in inp:
                job[fieldname] = copy.deepcopy(inp["default"])
            elif job.get(fieldname) is None and "null" in aslist(inp["type"]):
                job[fieldname] = None
            else:
                raise WorkflowException(
                    "Missing required input parameter '%s'" % shortname(cast(str, inp["id"]))
                )


def avroize_type(
    field_type: Union[CWLObjectType, MutableSequence[Any], CWLOutputType, None],
    name_prefix: str = "",
) -> Union[CWLObjectType, MutableSequence[Any], CWLOutputType, None]:
    """Add missing information to a type so that CWL types are valid."""
    if isinstance(field_type, MutableSequence):
        for i, field in enumerate(field_type):
            field_type[i] = avroize_type(field, name_prefix)
    elif isinstance(field_type, MutableMapping):
        if field_type["type"] in ("enum", "record"):
            if "name" not in field_type:
                field_type["name"] = name_prefix + str(uuid.uuid4())
        if field_type["type"] == "record":
            field_type["fields"] = avroize_type(
                cast(MutableSequence[CWLOutputType], field_type["fields"]), name_prefix
            )
        elif field_type["type"] == "array":
            field_type["items"] = avroize_type(
                cast(MutableSequence[CWLOutputType], field_type["items"]), name_prefix
            )
        else:
            field_type["type"] = avroize_type(cast(CWLOutputType, field_type["type"]), name_prefix)
    elif field_type == "File":
        return "org.w3id.cwl.cwl.File"
    elif field_type == "Directory":
        return "org.w3id.cwl.cwl.Directory"
    return field_type


def get_overrides(overrides: MutableSequence[CWLObjectType], toolid: str) -> CWLObjectType:
    """Combine overrides for the target tool ID."""
    req: CWLObjectType = {}
    if not isinstance(overrides, MutableSequence):
        raise ValidationException("Expected overrides to be a list, but was %s" % type(overrides))
    for ov in overrides:
        if ov["overrideTarget"] == toolid:
            req.update(ov)
    return req


_VAR_SPOOL_ERROR = textwrap.dedent(
    """
    Non-portable reference to /var/spool/cwl detected: '{}'.
    To fix, replace /var/spool/cwl with $(runtime.outdir) or add
    DockerRequirement to the 'requirements' section and declare
    'dockerOutputDirectory: /var/spool/cwl'.
    """
)


def var_spool_cwl_detector(
    obj: CWLOutputType,
    item: Optional[Any] = None,
    obj_key: Optional[Any] = None,
) -> bool:
    """Detect any textual reference to /var/spool/cwl."""
    r = False
    if isinstance(obj, str):
        if "var/spool/cwl" in obj and obj_key != "dockerOutputDirectory":
            _logger.warning(
                SourceLine(item=item, key=obj_key, raise_type=str).makeError(
                    _VAR_SPOOL_ERROR.format(obj)
                )
            )
            r = True
    elif isinstance(obj, MutableMapping):
        for mkey, mvalue in obj.items():
            r = var_spool_cwl_detector(cast(CWLOutputType, mvalue), obj, mkey) or r
    elif isinstance(obj, MutableSequence):
        for lkey, lvalue in enumerate(obj):
            r = var_spool_cwl_detector(cast(CWLOutputType, lvalue), obj, lkey) or r
    return r


def eval_resource(
    builder: Builder, resource_req: Union[str, int, float]
) -> Optional[Union[str, int, float]]:
    if isinstance(resource_req, str) and expression.needs_parsing(resource_req):
        result = builder.do_eval(resource_req)
        if isinstance(result, (str, int)) or result is None:
            return result
        raise WorkflowException(
            "Got incorrect return type {} from resource expression evaluation of {}.".format(
                type(result), resource_req
            )
        )
    return resource_req


# Threshold where the "too many files" warning kicks in
FILE_COUNT_WARNING = 5000


@mypyc_attr(allow_interpreted_subclasses=True)
class Process(HasReqsHints, metaclass=abc.ABCMeta):
    """Abstract CWL Process."""

    def __init__(self, toolpath_object: CommentedMap, loadingContext: LoadingContext) -> None:
        """Build a Process object from the provided dictionary."""
        super().__init__()
        self.metadata: CWLObjectType = getdefault(loadingContext.metadata, {})
        self.provenance_object: Optional["ProvenanceProfile"] = None
        self.parent_wf: Optional["ProvenanceProfile"] = None
        global SCHEMA_FILE, SCHEMA_DIR, SCHEMA_ANY  # pylint: disable=global-statement
        if SCHEMA_FILE is None or SCHEMA_ANY is None or SCHEMA_DIR is None:
            get_schema("v1.0")
            SCHEMA_ANY = cast(
                CWLObjectType,
                SCHEMA_CACHE["v1.0"][3].idx["https://w3id.org/cwl/salad#Any"],
            )
            SCHEMA_FILE = cast(
                CWLObjectType,
                SCHEMA_CACHE["v1.0"][3].idx["https://w3id.org/cwl/cwl#File"],
            )
            SCHEMA_DIR = cast(
                CWLObjectType,
                SCHEMA_CACHE["v1.0"][3].idx["https://w3id.org/cwl/cwl#Directory"],
            )

        self.names = make_avro_schema([SCHEMA_FILE, SCHEMA_DIR, SCHEMA_ANY], Loader({}))
        self.tool = toolpath_object
        debug = loadingContext.debug
        self.requirements = copy.deepcopy(getdefault(loadingContext.requirements, []))
        tool_requirements = self.tool.get("requirements", [])
        if tool_requirements is None:
            raise SourceLine(self.tool, "requirements", ValidationException, debug).makeError(
                "If 'requirements' is present then it must be a list "
                "or map/dictionary, not empty."
            )
        self.requirements.extend(tool_requirements)
        if "id" not in self.tool:
            self.tool["id"] = "_:" + str(uuid.uuid4())
        self.requirements.extend(
            cast(
                List[CWLObjectType],
                get_overrides(getdefault(loadingContext.overrides_list, []), self.tool["id"]).get(
                    "requirements", []
                ),
            )
        )
        self.hints = copy.deepcopy(getdefault(loadingContext.hints, []))
        tool_hints = self.tool.get("hints", [])
        if tool_hints is None:
            raise SourceLine(self.tool, "hints", ValidationException, debug).makeError(
                "If 'hints' is present then it must be a list " "or map/dictionary, not empty."
            )
        self.hints.extend(tool_hints)
        # Versions of requirements and hints which aren't mutated.
        self.original_requirements = copy.deepcopy(self.requirements)
        self.original_hints = copy.deepcopy(self.hints)
        self.doc_loader = loadingContext.loader
        self.doc_schema = loadingContext.avsc_names

        self.formatgraph: Optional[Graph] = None
        if self.doc_loader is not None:
            self.formatgraph = self.doc_loader.graph

        self.checkRequirements(self.tool, supportedProcessRequirements)
        self.validate_hints(
            cast(Names, loadingContext.avsc_names),
            self.tool.get("hints", []),
            strict=getdefault(loadingContext.strict, False),
        )

        self.schemaDefs: MutableMapping[str, CWLObjectType] = {}

        sd, _ = self.get_requirement("SchemaDefRequirement")

        if sd is not None:
            sdtypes = copy.deepcopy(cast(MutableSequence[CWLObjectType], sd["types"]))
            avroize_type(cast(MutableSequence[CWLOutputType], sdtypes))
            av = make_valid_avro(
                sdtypes,
                {cast(str, t["name"]): cast(Dict[str, Any], t) for t in sdtypes},
                set(),
                vocab=INPUT_OBJ_VOCAB,
            )
            for i in av:
                self.schemaDefs[i["name"]] = i  # type: ignore
            make_avsc_object(convert_to_dict(av), self.names)

        # Build record schema from inputs
        self.inputs_record_schema: CWLObjectType = {
            "name": "input_record_schema",
            "type": "record",
            "fields": [],
        }
        self.outputs_record_schema: CWLObjectType = {
            "name": "outputs_record_schema",
            "type": "record",
            "fields": [],
        }

        for key in ("inputs", "outputs"):
            for i in self.tool[key]:
                c = copy.deepcopy(i)
                c["name"] = shortname(c["id"])
                del c["id"]

                if "type" not in c:
                    raise ValidationException("Missing 'type' in parameter '{}'".format(c["name"]))

                if "default" in c and "null" not in aslist(c["type"]):
                    nullable = ["null"]
                    nullable.extend(aslist(c["type"]))
                    c["type"] = nullable
                else:
                    c["type"] = c["type"]

                c["type"] = avroize_type(c["type"], c["name"])
                if key == "inputs":
                    cast(List[CWLObjectType], self.inputs_record_schema["fields"]).append(c)
                elif key == "outputs":
                    cast(List[CWLObjectType], self.outputs_record_schema["fields"]).append(c)

        with SourceLine(toolpath_object, "inputs", ValidationException, debug):
            self.inputs_record_schema = cast(
                CWLObjectType,
                make_valid_avro(self.inputs_record_schema, {}, set()),
            )
            make_avsc_object(convert_to_dict(self.inputs_record_schema), self.names)
        with SourceLine(toolpath_object, "outputs", ValidationException, debug):
            self.outputs_record_schema = cast(
                CWLObjectType,
                make_valid_avro(self.outputs_record_schema, {}, set()),
            )
            make_avsc_object(convert_to_dict(self.outputs_record_schema), self.names)

        self.container_engine = "docker"
        if loadingContext.podman:
            self.container_engine = "podman"
        elif loadingContext.singularity:
            self.container_engine = "singularity"

        if toolpath_object.get("class") is not None and not getdefault(
            loadingContext.disable_js_validation, False
        ):
            validate_js_options: Optional[Dict[str, Union[List[str], str, int]]] = None
            if loadingContext.js_hint_options_file is not None:
                try:
                    with open(loadingContext.js_hint_options_file) as options_file:
                        validate_js_options = json.load(options_file)
                except (OSError, ValueError):
                    _logger.error(
                        "Failed to read options file %s",
                        loadingContext.js_hint_options_file,
                    )
                    raise
            if self.doc_schema is not None:
                classname = toolpath_object["class"]
                avroname = classname
                if self.doc_loader and classname in self.doc_loader.vocab:
                    avroname = avro_type_name(self.doc_loader.vocab[classname])
                validate_js_expressions(
                    toolpath_object,
                    self.doc_schema.names[avroname],
                    validate_js_options,
                    self.container_engine,
                    loadingContext.eval_timeout,
                )

        dockerReq, is_req = self.get_requirement("DockerRequirement")

        if (
            dockerReq is not None
            and "dockerOutputDirectory" in dockerReq
            and is_req is not None
            and not is_req
        ):
            _logger.warning(
                SourceLine(item=dockerReq, raise_type=str).makeError(
                    "When 'dockerOutputDirectory' is declared, DockerRequirement "
                    "should go in the 'requirements' section, not 'hints'."
                    ""
                )
            )

        if (
            dockerReq is not None
            and is_req is not None
            and dockerReq.get("dockerOutputDirectory") == "/var/spool/cwl"
        ):
            if is_req:
                # In this specific case, it is legal to have /var/spool/cwl, so skip the check.
                pass
            else:
                # Must be a requirement
                var_spool_cwl_detector(self.tool)
        else:
            var_spool_cwl_detector(self.tool)

    def _init_job(self, joborder: CWLObjectType, runtime_context: RuntimeContext) -> Builder:
        if self.metadata.get("cwlVersion") != INTERNAL_VERSION:
            raise WorkflowException(
                "Process object loaded with version '%s', must update to '%s' in order to execute."
                % (self.metadata.get("cwlVersion"), INTERNAL_VERSION)
            )

        job = copy.deepcopy(joborder)

        make_fs_access = getdefault(runtime_context.make_fs_access, StdFsAccess)
        fs_access = make_fs_access(runtime_context.basedir)

        load_listing_req, _ = self.get_requirement("LoadListingRequirement")

        load_listing = (
            cast(LoadListingType, load_listing_req.get("loadListing"))
            if load_listing_req is not None
            else "no_listing"
        )

        # Validate job order
        try:
            fill_in_defaults(self.tool["inputs"], job, fs_access)

            normalizeFilesDirs(job)
            schema = self.names.get_name("input_record_schema", None)
            if schema is None:
                raise WorkflowException("Missing input record schema: " "{}".format(self.names))
            validate_ex(
                schema,
                job,
                strict=False,
                logger=_logger_validation_warnings,
                vocab=INPUT_OBJ_VOCAB,
            )

            if load_listing != "no_listing":
                get_listing(fs_access, job, recursive=(load_listing == "deep_listing"))

            visit_class(job, ("File",), functools.partial(add_sizes, fs_access))

            if load_listing == "deep_listing":
                for i, inparm in enumerate(self.tool["inputs"]):
                    k = shortname(inparm["id"])
                    if k not in job:
                        continue
                    v = job[k]
                    dircount = [0]

                    def inc(d: List[int]) -> None:
                        d[0] += 1

                    visit_class(v, ("Directory",), lambda x: inc(dircount))  # noqa: B023
                    if dircount[0] == 0:
                        continue
                    filecount = [0]
                    visit_class(v, ("File",), lambda x: inc(filecount))  # noqa: B023
                    if filecount[0] > FILE_COUNT_WARNING:
                        # Long lines in this message are okay, will be reflowed based on terminal columns.
                        _logger.warning(
                            strip_dup_lineno(
                                SourceLine(self.tool["inputs"], i, str).makeError(
                                    "Recursive directory listing has resulted "
                                    "in a large number of File objects (%s) passed "
                                    "to the input parameter '%s'.  This may "
                                    "negatively affect workflow performance and memory use.\n\n"
                                    "If this is a problem, use the hint 'cwltool:LoadListingRequirement' "
                                    'with "shallow_listing" or "no_listing" to change the directory '
                                    """listing behavior:

$namespaces:
  cwltool: "http://commonwl.org/cwltool#"
hints:
  cwltool:LoadListingRequirement:
    loadListing: shallow_listing

"""
                                    % (filecount[0], k)
                                )
                            )
                        )

        except (ValidationException, WorkflowException) as err:
            raise WorkflowException("Invalid job input record:\n" + str(err)) from err

        files: List[CWLObjectType] = []
        bindings = CommentedSeq()
        outdir = ""
        tmpdir = ""
        stagedir = ""

        docker_req, _ = self.get_requirement("DockerRequirement")
        default_docker = None

        if docker_req is None and runtime_context.default_container:
            default_docker = runtime_context.default_container

        if (docker_req or default_docker) and runtime_context.use_container:
            if docker_req is not None:
                # Check if docker output directory is absolute
                if docker_req.get("dockerOutputDirectory") and cast(
                    str, docker_req.get("dockerOutputDirectory")
                ).startswith("/"):
                    outdir = cast(str, docker_req.get("dockerOutputDirectory"))
                else:
                    outdir = cast(
                        str,
                        docker_req.get("dockerOutputDirectory")
                        or runtime_context.docker_outdir
                        or random_outdir(),
                    )
            elif default_docker is not None:
                outdir = runtime_context.docker_outdir or random_outdir()
            tmpdir = runtime_context.docker_tmpdir or "/tmp"  # nosec
            stagedir = runtime_context.docker_stagedir or "/var/lib/cwl"
        else:
            if self.tool["class"] == "CommandLineTool":
                outdir = fs_access.realpath(runtime_context.get_outdir())
                tmpdir = fs_access.realpath(runtime_context.get_tmpdir())
                stagedir = fs_access.realpath(runtime_context.get_stagedir())

        cwl_version = cast(
            str,
            self.metadata.get(ORIGINAL_CWLVERSION, None),
        )
        builder = Builder(
            job,
            files,
            bindings,
            self.schemaDefs,
            self.names,
            self.requirements,
            self.hints,
            {},
            runtime_context.mutation_manager,
            self.formatgraph,
            make_fs_access,
            fs_access,
            runtime_context.job_script_provider,
            runtime_context.eval_timeout,
            runtime_context.debug,
            runtime_context.js_console,
            runtime_context.force_docker_pull,
            load_listing,
            outdir,
            tmpdir,
            stagedir,
            cwl_version,
            self.container_engine,
        )

        bindings.extend(
            builder.bind_input(
                self.inputs_record_schema,
                job,
                discover_secondaryFiles=getdefault(runtime_context.toplevel, False),
            )
        )

        if self.tool.get("baseCommand"):
            for index, command in enumerate(aslist(self.tool["baseCommand"])):
                bindings.append({"position": [-1000000, index], "datum": command})

        if self.tool.get("arguments"):
            for i, arg in enumerate(self.tool["arguments"]):
                lc = self.tool["arguments"].lc.data[i]
                filename = self.tool["arguments"].lc.filename
                bindings.lc.add_kv_line_col(len(bindings), lc)
                if isinstance(arg, MutableMapping):
                    arg = copy.deepcopy(arg)
                    if arg.get("position"):
                        position = arg.get("position")
                        if isinstance(position, str):  # no need to test the
                            # CWLVersion as the v1.0
                            # schema only allows ints
                            position = builder.do_eval(position)
                            if position is None:
                                position = 0
                        arg["position"] = [position, i]
                    else:
                        arg["position"] = [0, i]
                    bindings.append(arg)
                elif ("$(" in arg) or ("${" in arg):
                    cm = CommentedMap((("position", [0, i]), ("valueFrom", arg)))
                    cm.lc.add_kv_line_col("valueFrom", lc)
                    cm.lc.filename = filename
                    bindings.append(cm)
                else:
                    cm = CommentedMap((("position", [0, i]), ("datum", arg)))
                    cm.lc.add_kv_line_col("datum", lc)
                    cm.lc.filename = filename
                    bindings.append(cm)

        # use python2 like sorting of heterogeneous lists
        # (containing str and int types),
        key = functools.cmp_to_key(cmp_like_py2)

        # This awkward construction replaces the contents of
        # "bindings" in place (because Builder expects it to be
        # mutated in place, sigh, I'm sorry) with its contents sorted,
        # supporting different versions of Python and ruamel.yaml with
        # different behaviors/bugs in CommentedSeq.
        bindings_copy = copy.deepcopy(bindings)
        del bindings[:]
        bindings.extend(sorted(bindings_copy, key=key))

        if self.tool["class"] != "Workflow":
            builder.resources = self.evalResources(builder, runtime_context)
        return builder

    def evalResources(
        self, builder: Builder, runtimeContext: RuntimeContext
    ) -> Dict[str, Union[int, float]]:
        resourceReq, _ = self.get_requirement("ResourceRequirement")
        if resourceReq is None:
            resourceReq = {}

        cwl_version = self.metadata.get(ORIGINAL_CWLVERSION, None)
        if cwl_version == "v1.0":
            ram = 1024
        else:
            ram = 256
        request: Dict[str, Union[int, float, str]] = {
            "coresMin": 1,
            "coresMax": 1,
            "ramMin": ram,
            "ramMax": ram,
            "tmpdirMin": 1024,
            "tmpdirMax": 1024,
            "outdirMin": 1024,
            "outdirMax": 1024,
        }

        cudaReq, _ = self.get_requirement("http://commonwl.org/cwltool#CUDARequirement")
        if cudaReq:
            request["cudaDeviceCountMin"] = 1
            request["cudaDeviceCountMax"] = 1

        for rsc, a in (
            (resourceReq, "cores"),
            (resourceReq, "ram"),
            (resourceReq, "tmpdir"),
            (resourceReq, "outdir"),
            (cudaReq, "cudaDeviceCount"),
        ):
            if rsc is None:
                continue
            mn = mx = None  # type: Optional[Union[int, float]]
            if rsc.get(a + "Min"):
                mn = cast(
                    Union[int, float],
                    eval_resource(builder, cast(Union[str, int, float], rsc[a + "Min"])),
                )
            if rsc.get(a + "Max"):
                mx = cast(
                    Union[int, float],
                    eval_resource(builder, cast(Union[str, int, float], rsc[a + "Max"])),
                )
            if mn is None:
                mn = mx
            elif mx is None:
                mx = mn

            if mn is not None:
                request[a + "Min"] = mn
                request[a + "Max"] = cast(Union[int, float], mx)

        request_evaluated = cast(Dict[str, Union[int, float]], request)
        if runtimeContext.select_resources is not None:
            # Call select resources hook
            return runtimeContext.select_resources(request_evaluated, runtimeContext)

        defaultReq = {
            "cores": request_evaluated["coresMin"],
            "ram": math.ceil(request_evaluated["ramMin"]),
            "tmpdirSize": math.ceil(request_evaluated["tmpdirMin"]),
            "outdirSize": math.ceil(request_evaluated["outdirMin"]),
        }
        if cudaReq:
            defaultReq["cudaDeviceCount"] = request_evaluated["cudaDeviceCountMin"]
        return defaultReq

    def checkRequirements(
        self,
        rec: Union[MutableSequence[CWLObjectType], CWLObjectType, CWLOutputType, None],
        supported_process_requirements: Iterable[str],
    ) -> None:
        """Check the presence of unsupported requirements."""
        if isinstance(rec, MutableMapping):
            if "requirements" in rec:
                debug = _logger.isEnabledFor(logging.DEBUG)
                for i, entry in enumerate(
                    cast(MutableSequence[CWLObjectType], rec["requirements"])
                ):
                    with SourceLine(rec["requirements"], i, UnsupportedRequirement, debug):
                        if cast(str, entry["class"]) not in supported_process_requirements:
                            raise UnsupportedRequirement(
                                f"Unsupported requirement {entry['class']}."
                            )

    def validate_hints(self, avsc_names: Names, hints: List[CWLObjectType], strict: bool) -> None:
        """Process the hints field."""
        if self.doc_loader is None:
            return
        debug = _logger.isEnabledFor(logging.DEBUG)
        for i, r in enumerate(hints):
            sl = SourceLine(hints, i, ValidationException, debug)
            with sl:
                classname = cast(str, r["class"])
                if classname == "http://commonwl.org/cwltool#Loop":
                    raise ValidationException(
                        "http://commonwl.org/cwltool#Loop is valid only under requirements."
                    )
                avroname = classname
                if classname in self.doc_loader.vocab:
                    avroname = avro_type_name(self.doc_loader.vocab[classname])
                if avsc_names.get_name(avroname, None) is not None:
                    plain_hint = {
                        key: r[key] for key in r if key not in self.doc_loader.identifiers
                    }  # strip identifiers
                    validate_ex(
                        cast(
                            Schema,
                            avsc_names.get_name(avroname, None),
                        ),
                        plain_hint,
                        strict=strict,
                        vocab=self.doc_loader.vocab,
                    )
                elif r["class"] in ("NetworkAccess", "LoadListingRequirement"):
                    pass
                else:
                    _logger.info(str(sl.makeError("Unknown hint %s" % (r["class"]))))

    def visit(self, op: Callable[[CommentedMap], None]) -> None:
        op(self.tool)

    @abc.abstractmethod
    def job(
        self,
        job_order: CWLObjectType,
        output_callbacks: Optional[OutputCallbackType],
        runtimeContext: RuntimeContext,
    ) -> JobsGeneratorType:
        pass

    def __str__(self) -> str:
        """Return the id of this CWL process."""
        return f"{type(self).__name__}: {self.tool['id']}"


_names: Set[str] = set()


def uniquename(stem: str, names: Optional[Set[str]] = None) -> str:
    global _names
    if names is None:
        names = _names
    c = 1
    u = stem
    while u in names:
        c += 1
        u = f"{stem}_{c}"
    names.add(u)
    return u


def nestdir(base: str, deps: CWLObjectType) -> CWLObjectType:
    dirname = os.path.dirname(base) + "/"
    subid = cast(str, deps["location"])
    if subid.startswith(dirname):
        s2 = subid[len(dirname) :]
        sp = s2.split("/")
        sp.pop()
        while sp:
            loc = dirname + "/".join(sp)
            nx = sp.pop()
            deps = {
                "class": "Directory",
                "basename": nx,
                "listing": [deps],
                "location": loc,
            }
    return deps


def mergedirs(
    listing: MutableSequence[CWLObjectType],
) -> MutableSequence[CWLObjectType]:
    r: List[CWLObjectType] = []
    ents: Dict[str, CWLObjectType] = {}
    for e in listing:
        basename = cast(str, e["basename"])
        if basename not in ents:
            ents[basename] = e
        elif e["location"] != ents[basename]["location"]:
            raise ValidationException(
                "Conflicting basename in listing or secondaryFiles, '%s' used by both '%s' and '%s'"
                % (basename, e["location"], ents[basename]["location"])
            )
        elif e["class"] == "Directory":
            if e.get("listing"):
                # name already in entries
                # merge it into the existing listing
                cast(List[CWLObjectType], ents[basename].setdefault("listing", [])).extend(
                    cast(List[CWLObjectType], e["listing"])
                )
    for e in ents.values():
        if e["class"] == "Directory" and "listing" in e:
            e["listing"] = cast(
                MutableSequence[CWLOutputAtomType],
                mergedirs(cast(List[CWLObjectType], e["listing"])),
            )
    r.extend(ents.values())
    return r


CWL_IANA = "https://www.iana.org/assignments/media-types/application/cwl"


def scandeps(
    base: str,
    doc: Union[CWLObjectType, MutableSequence[CWLObjectType]],
    reffields: Set[str],
    urlfields: Set[str],
    loadref: Callable[[str, str], Union[CommentedMap, CommentedSeq, str, None]],
    urljoin: Callable[[str, str], str] = urllib.parse.urljoin,
    nestdirs: bool = True,
) -> MutableSequence[CWLObjectType]:
    """
    Search for external files references in a CWL document or input object.

    Looks for objects with 'class: File' or 'class: Directory' and
    adds them to the list of dependencies.

    :param base: the base URL for relative references.
    :param doc: a CWL document or input object
    :param urlfields: added as a File dependency
    :param reffields: field name like a workflow step 'run'; will be
      added as a dependency and also loaded (using the 'loadref'
      function) and recursively scanned for dependencies.  Those
      dependencies will be added as secondary files to the primary file.
    :param nestdirs: if true, create intermediate directory objects when
      a file is located in a subdirectory under the starting directory.
      This is so that if the dependencies are materialized, they will
      produce the same relative file system locations.
    :returns: A list of File or Directory dependencies
    """
    r: MutableSequence[CWLObjectType] = []
    if isinstance(doc, MutableMapping):
        if "id" in doc:
            if cast(str, doc["id"]).startswith("file://"):
                df, _ = urllib.parse.urldefrag(cast(str, doc["id"]))
                if base != df:
                    r.append({"class": "File", "location": df, "format": CWL_IANA})
                    base = df

        if doc.get("class") in ("File", "Directory") and "location" in urlfields:
            u = cast(Optional[str], doc.get("location", doc.get("path")))
            if u and not u.startswith("_:"):
                deps: CWLObjectType = {
                    "class": doc["class"],
                    "location": urljoin(base, u),
                }
                if "basename" in doc:
                    deps["basename"] = doc["basename"]
                if doc["class"] == "Directory" and "listing" in doc:
                    deps["listing"] = doc["listing"]
                if doc["class"] == "File" and "secondaryFiles" in doc:
                    deps["secondaryFiles"] = cast(
                        CWLOutputAtomType,
                        scandeps(
                            base,
                            cast(
                                Union[CWLObjectType, MutableSequence[CWLObjectType]],
                                doc["secondaryFiles"],
                            ),
                            reffields,
                            urlfields,
                            loadref,
                            urljoin=urljoin,
                            nestdirs=nestdirs,
                        ),
                    )
                if nestdirs:
                    deps = nestdir(base, deps)
                r.append(deps)
            else:
                if doc["class"] == "Directory" and "listing" in doc:
                    r.extend(
                        scandeps(
                            base,
                            cast(MutableSequence[CWLObjectType], doc["listing"]),
                            reffields,
                            urlfields,
                            loadref,
                            urljoin=urljoin,
                            nestdirs=nestdirs,
                        )
                    )
                elif doc["class"] == "File" and "secondaryFiles" in doc:
                    r.extend(
                        scandeps(
                            base,
                            cast(MutableSequence[CWLObjectType], doc["secondaryFiles"]),
                            reffields,
                            urlfields,
                            loadref,
                            urljoin=urljoin,
                            nestdirs=nestdirs,
                        )
                    )

        for k, v in doc.items():
            if k in reffields:
                for u2 in aslist(v):
                    if isinstance(u2, MutableMapping):
                        r.extend(
                            scandeps(
                                base,
                                u2,
                                reffields,
                                urlfields,
                                loadref,
                                urljoin=urljoin,
                                nestdirs=nestdirs,
                            )
                        )
                    else:
                        subid = urljoin(base, u2)
                        basedf, _ = urllib.parse.urldefrag(base)
                        subiddf, _ = urllib.parse.urldefrag(subid)
                        if basedf == subiddf:
                            continue
                        sub = cast(
                            Union[MutableSequence[CWLObjectType], CWLObjectType],
                            loadref(base, u2),
                        )
                        deps2: CWLObjectType = {
                            "class": "File",
                            "location": subid,
                            "format": CWL_IANA,
                        }
                        sf = scandeps(
                            subid,
                            sub,
                            reffields,
                            urlfields,
                            loadref,
                            urljoin=urljoin,
                            nestdirs=nestdirs,
                        )
                        if sf:
                            deps2["secondaryFiles"] = cast(
                                MutableSequence[CWLOutputAtomType], mergedirs(sf)
                            )
                        if nestdirs:
                            deps2 = nestdir(base, deps2)
                        r.append(deps2)
            elif k in urlfields and k != "location":
                for u3 in aslist(v):
                    deps = {"class": "File", "location": urljoin(base, u3)}
                    if nestdirs:
                        deps = nestdir(base, deps)
                    r.append(deps)
            elif doc.get("class") in ("File", "Directory") and k in (
                "listing",
                "secondaryFiles",
            ):
                # should be handled earlier.
                pass
            else:
                r.extend(
                    scandeps(
                        base,
                        cast(Union[MutableSequence[CWLObjectType], CWLObjectType], v),
                        reffields,
                        urlfields,
                        loadref,
                        urljoin=urljoin,
                        nestdirs=nestdirs,
                    )
                )
    elif isinstance(doc, MutableSequence):
        for d in doc:
            r.extend(
                scandeps(
                    base,
                    d,
                    reffields,
                    urlfields,
                    loadref,
                    urljoin=urljoin,
                    nestdirs=nestdirs,
                )
            )

    if r:
        normalizeFilesDirs(r)

    return r


def compute_checksums(fs_access: StdFsAccess, fileobj: CWLObjectType) -> None:
    if "checksum" not in fileobj:
        checksum = hashlib.sha1()  # nosec
        location = cast(str, fileobj["location"])
        with fs_access.open(location, "rb") as f:
            contents = f.read(1024 * 1024)
            while contents != b"":
                checksum.update(contents)
                contents = f.read(1024 * 1024)
        fileobj["checksum"] = "sha1$%s" % checksum.hexdigest()
        fileobj["size"] = fs_access.size(location)



# Stubs for networkx.algorithms.connectivity.cuts (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.algorithms.flow import edmonds_karp

default_flow_func = edmonds_karp

def minimum_st_edge_cut(
    G,
    s,
    t,
    flow_func: Optional[Any] = ...,
    auxiliary: Optional[Any] = ...,
    residual: Optional[Any] = ...,
): ...
def minimum_st_node_cut(
    G,
    s,
    t,
    flow_func: Optional[Any] = ...,
    auxiliary: Optional[Any] = ...,
    residual: Optional[Any] = ...,
): ...
def minimum_node_cut(
    G, s: Optional[Any] = ..., t: Optional[Any] = ..., flow_func: Optional[Any] = ...
): ...
def minimum_edge_cut(
    G, s: Optional[Any] = ..., t: Optional[Any] = ..., flow_func: Optional[Any] = ...
): ...



# Stubs for networkx.algorithms.isomorphism.vf2userfunc (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from . import isomorphvf2 as vf2

class GraphMatcher(vf2.GraphMatcher):
    node_match: Any = ...
    edge_match: Any = ...
    G1_adj: Any = ...
    G2_adj: Any = ...
    def __init__(
        self, G1, G2, node_match: Optional[Any] = ..., edge_match: Optional[Any] = ...
    ) -> None: ...
    semantic_feasibility: Any = ...

class DiGraphMatcher(vf2.DiGraphMatcher):
    node_match: Any = ...
    edge_match: Any = ...
    G1_adj: Any = ...
    G2_adj: Any = ...
    def __init__(
        self, G1, G2, node_match: Optional[Any] = ..., edge_match: Optional[Any] = ...
    ) -> None: ...
    def semantic_feasibility(self, G1_node, G2_node): ...

class MultiGraphMatcher(GraphMatcher): ...
class MultiDiGraphMatcher(DiGraphMatcher): ...



# Stubs for networkx.algorithms.components (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .attracting import *
from .biconnected import *
from .connected import *
from .semiconnected import *
from .strongly_connected import *
from .weakly_connected import *



# Stubs for networkx.utils.decorators (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def not_implemented_for(*graph_types): ...
def open_file(path_arg, mode: str = ...): ...
def nodes_or_number(which_args): ...
def preserve_random_state(func): ...
def random_state(random_state_index): ...



import urllib.parse
from io import BytesIO
from pathlib import Path
from typing import IO, Any, List, cast

import pytest
from schema_salad.sourceline import cmap

from cwltool.command_line_tool import CommandLineTool
from cwltool.context import LoadingContext, RuntimeContext
from cwltool.main import main
from cwltool.stdfsaccess import StdFsAccess
from cwltool.update import INTERNAL_VERSION
from cwltool.utils import CWLObjectType
from ruamel.yaml.comments import CommentedMap

from .util import needs_docker

script = """
#!/usr/bin/env cwl-runner
cwlVersion: v1.0
class: CommandLineTool
inputs:
  - id: input
    type: File
    inputBinding:
      position: 0
  - id: output
    type: string
outputs:
  - id: output
    type: File
    outputBinding:
      glob: "$(inputs.output)"
stdout: "$(inputs.output)"
baseCommand: [cat]
"""


@needs_docker
def test_spaces_in_input_files(tmp_path: Path) -> None:
    script_name = tmp_path / "script"
    spaces = tmp_path / "test with spaces"
    spaces.touch()
    with script_name.open(mode="w") as script_file:
        script_file.write(script)

    params = [
        "--debug",
        "--outdir",
        str(tmp_path / "outdir"),
        str(script_name),
        "--input",
        str(spaces),
        "--output",
        "test.txt",
    ]
    assert main(params) == 1
    assert main(["--relax-path-checks"] + params) == 0


@needs_docker
@pytest.mark.parametrize(
    "filename", ["測試", "그래프", "график", "𒁃", "☕😍", "امتحان", "abc+DEFGZ.z_12345-"]
)
def test_unicode_in_input_files(tmp_path: Path, filename: str) -> None:
    script_name = tmp_path / "script"
    inputfile = tmp_path / filename
    inputfile.touch()
    with script_name.open(mode="w") as script_file:
        script_file.write(script)

    params = [
        "--debug",
        "--outdir",
        str(tmp_path / "outdir"),
        str(script_name),
        "--input",
        str(inputfile),
        "--output",
        "test.txt",
    ]
    assert main(params) == 0


@needs_docker
@pytest.mark.parametrize(
    "filename", ["測試", "그래프", "график", "𒁃", "☕😍", "امتحان", "abc+DEFGZ.z_12345-"]
)
def test_unicode_in_output_files(tmp_path: Path, filename: str) -> None:
    script_name = tmp_path / "script"
    inputfile = tmp_path / "test"
    inputfile.touch()
    with script_name.open(mode="w") as script_file:
        script_file.write(script)

    params = [
        "--debug",
        "--outdir",
        str(tmp_path / "outdir"),
        str(script_name),
        "--input",
        str(inputfile),
        "--output",
        filename,
    ]
    assert main(params) == 0


class TestFsAccess(StdFsAccess):
    """Stub fs access object that doesn't rely on the filesystem."""

    def glob(self, pattern: str) -> List[str]:
        """glob."""
        return [pattern]

    def open(self, fn: str, mode: str) -> IO[Any]:
        """open."""
        return BytesIO(b"aoeu")

    def isfile(self, fn: str) -> bool:
        """isfile."""
        return True

    def size(self, fn: str) -> int:
        """size."""
        return 4


def test_clt_returns_specialchar_names(tmp_path: Path) -> None:
    """Confirm that special characters in filenames do not cause problems."""
    loading_context = LoadingContext(
        {
            "metadata": {
                "cwlVersion": INTERNAL_VERSION,
                "http://commonwl.org/cwltool#original_cwlVersion": INTERNAL_VERSION,
            }
        }
    )
    clt = CommandLineTool(
        cast(
            CommentedMap,
            cmap(
                {
                    "cwlVersion": INTERNAL_VERSION,
                    "class": "CommandLineTool",
                    "inputs": [],
                    "outputs": [],
                    "requirements": [],
                }
            ),
        ),
        loading_context,
    )

    # Reserved characters will be URL encoded during the creation of a file URI
    # Internal references to files are in URI form, and are therefore URL encoded
    # Final output files should not retain their URL encoded filenames
    rfc_3986_gen_delims = [":", "/", "?", "#", "[", "]", "@"]
    rfc_3986_sub_delims = ["!", "$", "&", "'", "(", ")", "*", "+", ",", ";", "="]
    unix_reserved = ["/", "\0"]
    reserved = [
        special_char
        for special_char in (rfc_3986_gen_delims + rfc_3986_sub_delims)
        if special_char not in unix_reserved
    ]

    # Mock an "output" file with the above special characters in its name
    special = "".join(reserved)
    output_schema = cast(CWLObjectType, {"type": "File", "outputBinding": {"glob": special}})
    mock_output = tmp_path / special
    mock_output.touch()

    # Prepare minimal arguments for CommandLineTool.collect_output()
    builder = clt._init_job({}, RuntimeContext())
    builder.pathmapper = clt.make_path_mapper(
        builder.files, builder.stagedir, RuntimeContext(), True
    )
    fs_access = builder.make_fs_access(str(tmp_path))

    result = cast(
        CWLObjectType,
        clt.collect_output(output_schema, builder, str(tmp_path), fs_access),
    )

    assert result["class"] == "File"
    assert result["basename"] == special
    assert result["nameroot"] == special
    assert str(result["location"]).endswith(urllib.parse.quote(special))

    # Now test when outdir is a URI, make sure it doesn't get
    # incorrectly quoted as a file.
    builder = clt._init_job({}, RuntimeContext())
    builder.pathmapper = clt.make_path_mapper(
        builder.files, builder.stagedir, RuntimeContext(), True
    )
    builder.outdir = "/var/spool/cwl"
    fs_access = TestFsAccess("")

    result = cast(
        CWLObjectType,
        clt.collect_output(
            output_schema,
            builder,
            "keep:ae755cd1b3cff63152ff4200f4dea7e9+52",
            fs_access,
        ),
    )

    assert result["class"] == "File"
    assert result["basename"] == special
    assert result["nameroot"] == special
    assert (
        result["location"]
        == "keep:ae755cd1b3cff63152ff4200f4dea7e9+52/%3A%3F%23%5B%5D%40%21%24%26%27%28%29%2A%2B%2C%3B%3D"
    )



# Stubs for networkx.algorithms.link_analysis.pagerank_alg (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def pagerank(
    G,
    alpha: float = ...,
    personalization: Optional[Any] = ...,
    max_iter: int = ...,
    tol: float = ...,
    nstart: Optional[Any] = ...,
    weight: str = ...,
    dangling: Optional[Any] = ...,
): ...
def google_matrix(
    G,
    alpha: float = ...,
    personalization: Optional[Any] = ...,
    nodelist: Optional[Any] = ...,
    weight: str = ...,
    dangling: Optional[Any] = ...,
): ...
def pagerank_numpy(
    G,
    alpha: float = ...,
    personalization: Optional[Any] = ...,
    weight: str = ...,
    dangling: Optional[Any] = ...,
): ...
def pagerank_scipy(
    G,
    alpha: float = ...,
    personalization: Optional[Any] = ...,
    max_iter: int = ...,
    tol: float = ...,
    weight: str = ...,
    dangling: Optional[Any] = ...,
): ...



# Stubs for networkx.algorithms.tournament (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def is_tournament(G): ...
def hamiltonian_path(G): ...
def random_tournament(n): ...
def score_sequence(G): ...
def is_reachable(G, s, t): ...
def is_strongly_connected(G): ...



# Stubs for networkx.algorithms.isomorphism (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.isomorphism.isomorph import *
from networkx.algorithms.isomorphism.matchhelpers import *
from networkx.algorithms.isomorphism.temporalisomorphvf2 import *
from networkx.algorithms.isomorphism.vf2userfunc import *



# Stubs for networkx.algorithms.community.kclique (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def k_clique_communities(G, k, cliques: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.lowest_common_ancestors (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def tree_all_pairs_lowest_common_ancestor(
    G, root: Optional[Any] = ..., pairs: Optional[Any] = ...
): ...
def lowest_common_ancestor(G, node1, node2, default: Optional[Any] = ...): ...
def all_pairs_lowest_common_ancestor(G, pairs: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.isomorphism.temporalisomorphvf2 (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from .isomorphvf2 import DiGraphMatcher, GraphMatcher

class TimeRespectingGraphMatcher(GraphMatcher):
    temporal_attribute_name: Any = ...
    delta: Any = ...
    def __init__(self, G1, G2, temporal_attribute_name, delta) -> None: ...
    def one_hop(self, Gx, Gx_node, neighbors): ...
    def two_hop(self, Gx, core_x, Gx_node, neighbors): ...
    def semantic_feasibility(self, G1_node, G2_node): ...

class TimeRespectingDiGraphMatcher(DiGraphMatcher):
    temporal_attribute_name: Any = ...
    delta: Any = ...
    def __init__(self, G1, G2, temporal_attribute_name, delta) -> None: ...
    def get_pred_dates(self, Gx, Gx_node, core_x, pred): ...
    def get_succ_dates(self, Gx, Gx_node, core_x, succ): ...
    def one_hop(self, Gx, Gx_node, core_x, pred, succ): ...
    def two_hop_pred(self, Gx, Gx_node, core_x, pred): ...
    def two_hop_succ(self, Gx, Gx_node, core_x, succ): ...
    def preds(self, Gx, core_x, v, Gx_node: Optional[Any] = ...): ...
    def succs(self, Gx, core_x, v, Gx_node: Optional[Any] = ...): ...
    def test_one(self, pred_dates, succ_dates): ...
    def test_two(self, pred_dates, succ_dates): ...
    def semantic_feasibility(self, G1_node, G2_node): ...



# Stubs for networkx.algorithms.community.label_propagation (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def asyn_lpa_communities(G, weight: Optional[Any] = ...): ...
def label_propagation_communities(G): ...



# Stubs for networkx.readwrite.gml (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

long = int
unicode = str
unichr = chr

def read_gml(path, label: str = ..., destringizer: Optional[Any] = ...): ...
def parse_gml(lines, label: str = ..., destringizer: Optional[Any] = ...): ...
def generate_gml(G, stringizer: Optional[Any] = ...): ...
def write_gml(G, path, stringizer: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.assortativity (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.algorithms.assortativity.connectivity import *
from networkx.algorithms.assortativity.correlation import *
from networkx.algorithms.assortativity.mixing import *
from networkx.algorithms.assortativity.neighbor_degree import *
from networkx.algorithms.assortativity.pairs import *



# Stubs for networkx.algorithms.bridges (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def bridges(G, root: Optional[Any] = ...): ...
def has_bridges(G, root: Optional[Any] = ...): ...
def local_bridges(G, with_span: bool = ..., weight: Optional[Any] = ...): ...



# Stubs for networkx.algorithms.components.biconnected (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def is_biconnected(G): ...
def biconnected_component_edges(G): ...
def biconnected_components(G): ...
def biconnected_component_subgraphs(G, copy: bool = ...): ...
def articulation_points(G): ...



"""Tests --validation."""


from .util import get_data, get_main_output


def test_validate_graph_with_no_default() -> None:
    """Ensure that --validate works on $graph docs that lack a main/#main."""
    exit_code, stdout, stderr = get_main_output(
        ["--validate", get_data("tests/wf/packed_no_main.cwl")]
    )
    assert exit_code == 0
    assert "packed_no_main.cwl#echo is valid CWL" in stdout
    assert "packed_no_main.cwl#cat is valid CWL" in stdout
    assert "packed_no_main.cwl#collision is valid CWL" in stdout
    assert "tests/wf/packed_no_main.cwl is valid CWL" in stdout



# Stubs for networkx.version (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

version: str
date: str
dev: bool
version_info: Any
date_info: Any
vcs_info: Any



# Stubs for networkx.algorithms.dominating (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def dominating_set(G, start_with: Optional[Any] = ...): ...
def is_dominating_set(G, nbunch): ...



# Stubs for networkx.algorithms.centrality.betweenness_subset (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def betweenness_centrality_subset(
    G, sources, targets, normalized: bool = ..., weight: Optional[Any] = ...
): ...
def edge_betweenness_centrality_subset(
    G, sources, targets, normalized: bool = ..., weight: Optional[Any] = ...
): ...
def betweenness_centrality_source(
    G, normalized: bool = ..., weight: Optional[Any] = ..., sources: Optional[Any] = ...
): ...



# Stubs for networkx.algorithms.centrality.subgraph_alg (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from networkx.utils import *

def subgraph_centrality_exp(G): ...
def subgraph_centrality(G): ...
def communicability_betweenness_centrality(G, normalized: bool = ...): ...
def estrada_index(G): ...



# Stubs for networkx.algorithms.flow.maxflow (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from .preflowpush import preflow_push

default_flow_func = preflow_push

def maximum_flow(flowG, _s, _t, capacity: str = ..., flow_func: Optional[Any] = ..., **kwargs): ...
def maximum_flow_value(
    flowG, _s, _t, capacity: str = ..., flow_func: Optional[Any] = ..., **kwargs
): ...
def minimum_cut(flowG, _s, _t, capacity: str = ..., flow_func: Optional[Any] = ..., **kwargs): ...
def minimum_cut_value(
    flowG, _s, _t, capacity: str = ..., flow_func: Optional[Any] = ..., **kwargs
): ...



import json
from pathlib import Path
from typing import Any

from schema_salad.utils import convert_to_dict

from cwltool.context import LoadingContext
from cwltool.load_tool import load_tool
from cwltool.subgraph import get_step, get_subgraph
from cwltool.workflow import Workflow, default_make_tool

from .util import get_data, get_main_output, needs_docker


def clean(val: Any, path: str) -> Any:
    """Remove the path prefix from an string values."""
    if isinstance(val, str):
        if val.startswith(path):
            return val[len(path) + 1 :]
    if isinstance(val, dict):
        return {k: clean(v, path) for k, v in val.items()}
    if isinstance(val, list):
        return [clean(v, path) for v in val]
    return val


def test_get_subgraph() -> None:
    """Compare known correct subgraphs to generated subgraphs."""
    loading_context = LoadingContext({"construct_tool_object": default_make_tool})
    wf = Path(get_data("tests/subgraph/count-lines1-wf.cwl")).as_uri()
    loading_context.do_update = False
    tool = load_tool(wf, loading_context)

    sg = Path(get_data("tests/subgraph")).as_uri()

    for a in (
        "file1",
        "file2",
        "file3",
        "count_output",
        "output3",
        "output4",
        "output5",
        "step1",
        "step2",
        "step3",
        "step4",
        "step5",
    ):
        assert isinstance(tool, Workflow)
        extracted = get_subgraph([wf + "#" + a], tool, loading_context)
        with open(get_data("tests/subgraph/extract_" + a + ".json")) as f:
            assert json.load(f) == clean(convert_to_dict(extracted), sg)


def test_get_subgraph_long_out_form() -> None:
    """Compare subgraphs generatation when 'out' is in the long form."""
    loading_context = LoadingContext({"construct_tool_object": default_make_tool})
    wf = Path(get_data("tests/subgraph/1432.cwl")).as_uri()
    loading_context.do_update = False
    tool = load_tool(wf, loading_context)

    sg = Path(get_data("tests/")).as_uri()

    assert isinstance(tool, Workflow)
    extracted = get_subgraph([wf + "#step2"], tool, loading_context)
    with open(get_data("tests/subgraph/extract_step2_1432.json")) as f:
        assert json.load(f) == clean(convert_to_dict(extracted), sg)


def test_get_step() -> None:
    loading_context = LoadingContext({"construct_tool_object": default_make_tool})
    wf = Path(get_data("tests/subgraph/count-lines1-wf.cwl")).as_uri()
    loading_context.do_update = False
    tool = load_tool(wf, loading_context)
    assert isinstance(tool, Workflow)

    sg = Path(get_data("tests/subgraph")).as_uri()

    for a in (
        "step1",
        "step2",
        "step3",
        "step4",
        "step5",
    ):
        extracted = get_step(tool, wf + "#" + a, loading_context)
        with open(get_data("tests/subgraph/single_" + a + ".json")) as f:
            assert json.load(f) == clean(convert_to_dict(extracted), sg)


def test_single_process_inherit_reqshints(tmp_path: Path) -> None:
    """Inherit reqs and hints from parent(s) with --single-process."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$cdc1e84968261d6a7575b5305945471f8be199b6"


def test_single_process_inherit_hints_collision(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process hints collision."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_hint_collision.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$b3ec4ed1749c207e52b3a6d08c59f31d83bff519"


def test_single_process_inherit_reqs_collision(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process reqs collision."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_req_collision.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$b3ec4ed1749c207e52b3a6d08c59f31d83bff519"


def test_single_process_inherit_reqs_step_collision(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process reqs collision."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/steplevel-resreq.cwl"),
        ]
    )
    assert err_code == 0
    assert (
        json.loads(stdout)["output"]["checksum"] == "sha1$e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e"
    )


def test_single_process_inherit_reqs_hints_collision(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process reqs + hints collision."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_hint_req_collision.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$b3ec4ed1749c207e52b3a6d08c59f31d83bff519"


def test_single_process_inherit_only_hints(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process only hints."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_only_hint.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$ab5f2a9add5f54622dde555ac8ae9a3000e5ee0a"


def test_single_process_subwf_step(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process on sub-workflow step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "sub_wf/step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_subwf.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$cdc1e84968261d6a7575b5305945471f8be199b6"


def test_single_process_packed_subwf_step(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-process on packed sub-workflow step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "sub_wf/step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_subwf-packed.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$cdc1e84968261d6a7575b5305945471f8be199b6"


@needs_docker
def test_single_process_subwf_subwf_inline_step() -> None:
    """Test --single-process on an inline sub-sub-workflow step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-process",
            "step1/stepX/stepY",
            get_data("tests/subgraph/count-lines17-wf.cwl.json"),
            get_data("tests/wf/wc-job.json"),
        ]
    )
    assert err_code == 0
    assert (
        json.loads(stdout)["output"]["checksum"] == "sha1$3596ea087bfdaf52380eae441077572ed289d657"
    )


def test_single_step_subwf_step(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-step on sub-workflow step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-step",
            "sub_wf/step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_subwf.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$7608e5669ba454c61fab01c9b133b52a9a7de68c"


def test_single_step_wfstep_long_out(tmp_path: Path) -> None:
    """Support long form of step.out with --single-step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-step",
            "sub_wf/step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_subwf_b.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$7608e5669ba454c61fab01c9b133b52a9a7de68c"


def test_single_step_packed_subwf_step(tmp_path: Path) -> None:
    """Inherit reqs and hints --single-step on packed sub-workflow step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-step",
            "sub_wf/step1",
            "--outdir",
            str(tmp_path),
            get_data("tests/subgraph/env-wf2_subwf-packed.cwl"),
            get_data("tests/subgraph/env-job.json"),
        ]
    )
    assert err_code == 0
    assert json.loads(stdout)["out"]["checksum"] == "sha1$7608e5669ba454c61fab01c9b133b52a9a7de68c"


@needs_docker
def test_single_with_step_level_default_value() -> None:
    """Inherit step-level defaults with --single-step."""
    err_code, stdout, stderr = get_main_output(
        [
            "--single-step",
            "task2",
            get_data("tests/wf/cache_test_workflow.cwl"),
        ]
    )
    assert err_code == 0
    assert "two" in stderr


def test_print_targets_embedded_step() -> None:
    """Confirm that --print-targets works when a Workflow has embedded Processes."""
    err_code, stdout, stderr = get_main_output(
        [
            "--print-targets",
            get_data("tests/subgraph/timelimit2-wf.cwl"),
        ]
    )
    assert err_code == 0


def test_print_targets_embedded_reqsinherit() -> None:
    """Confirm --print-targets works with a step that needs a req from parent Workflow."""
    err_code, stdout, stderr = get_main_output(
        [
            "--print-targets",
            get_data("tests/wf/double-nested.cwl"),
        ]
    )
    assert err_code == 0


@needs_docker
def test_print_targets_embedded_sub_subwfs() -> None:
    """Confirm that --print-targets works with inline sub-sub-workflows."""
    err_code, stdout, stderr = get_main_output(
        [
            "--print-targets",
            get_data("tests/subgraph/count-lines17-wf.cwl.json"),
        ]
    )
    assert err_code == 0
    assert "step1/stepX/stepY" in stdout



# Stubs for networkx.readwrite.nx_shp (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def read_shp(path, simplify: bool = ..., geom_attrs: bool = ..., strict: bool = ...): ...
def write_shp(G, outdir): ...



# Stubs for networkx.generators.nonisomorphic_trees (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

def nonisomorphic_trees(order, create: str = ...): ...
def number_of_nonisomorphic_trees(order): ...



# Stubs for networkx.algorithms.operators.binary (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def union(G, H, rename: Any = ..., name: Optional[Any] = ...): ...
def disjoint_union(G, H): ...
def intersection(G, H): ...
def difference(G, H): ...
def symmetric_difference(G, H): ...
def compose(G, H): ...



# Stubs for networkx.exception (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

class NetworkXException(Exception): ...
class NetworkXError(NetworkXException): ...
class NetworkXPointlessConcept(NetworkXException): ...
class NetworkXAlgorithmError(NetworkXException): ...
class NetworkXUnfeasible(NetworkXAlgorithmError): ...
class NetworkXNoPath(NetworkXUnfeasible): ...
class NetworkXNoCycle(NetworkXUnfeasible): ...
class HasACycle(NetworkXException): ...
class NetworkXUnbounded(NetworkXAlgorithmError): ...
class NetworkXNotImplemented(NetworkXException): ...
class NodeNotFound(NetworkXException): ...
class AmbiguousSolution(NetworkXException): ...
class ExceededMaxIterations(NetworkXException): ...

class PowerIterationFailedConvergence(ExceededMaxIterations):
    def __init__(self, num_iterations, *args, **kw) -> None: ...



# Stubs for networkx.algorithms.link_prediction (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def resource_allocation_index(G, ebunch: Optional[Any] = ...): ...
def jaccard_coefficient(G, ebunch: Optional[Any] = ...): ...
def adamic_adar_index(G, ebunch: Optional[Any] = ...): ...
def preferential_attachment(G, ebunch: Optional[Any] = ...): ...
def cn_soundarajan_hopcroft(G, ebunch: Optional[Any] = ..., community: str = ...): ...
def ra_index_soundarajan_hopcroft(G, ebunch: Optional[Any] = ..., community: str = ...): ...
def within_inter_cluster(
    G, ebunch: Optional[Any] = ..., delta: float = ..., community: str = ...
): ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class RDF(DefinedNamespace):
    nil: URIRef
    direction: URIRef
    first: URIRef
    language: URIRef
    object: URIRef
    predicate: URIRef
    rest: URIRef
    subject: URIRef
    type: URIRef
    value: URIRef
    Alt: URIRef
    Bag: URIRef
    CompoundLiteral: URIRef
    List: URIRef
    Property: URIRef
    Seq: URIRef
    Statement: URIRef
    HTML: URIRef
    JSON: URIRef
    PlainLiteral: URIRef
    XMLLiteral: URIRef
    langString: URIRef



# Stubs for networkx.algorithms.flow.utils (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class CurrentEdge:
    def __init__(self, edges) -> None: ...
    def get(self): ...
    def move_to_next(self): ...

class Level:
    active: Any = ...
    inactive: Any = ...
    def __init__(self) -> None: ...

class GlobalRelabelThreshold:
    def __init__(self, n, m, freq) -> None: ...
    def add_work(self, work): ...
    def is_reached(self): ...
    def clear_work(self): ...

def build_residual_network(G, capacity): ...
def detect_unboundedness(R, s, t): ...
def build_flow_dict(G, R): ...



# Stubs for urllib.parse
import sys
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

_Str = Union[bytes, str]

uses_relative: List[str]
uses_netloc: List[str]
uses_params: List[str]
non_hierarchical: List[str]
uses_query: List[str]
uses_fragment: List[str]
scheme_chars: str
MAX_CACHE_SIZE = 0

class _ResultMixinBase(Generic[AnyStr]):
    def geturl(self) -> AnyStr: ...

class _ResultMixinStr(_ResultMixinBase[str]):
    def encode(self, encoding: str = ..., errors: str = ...) -> _ResultMixinBytes: ...

class _ResultMixinBytes(_ResultMixinBase[str]):
    def decode(self, encoding: str = ..., errors: str = ...) -> _ResultMixinStr: ...

class _NetlocResultMixinBase(Generic[AnyStr]):
    username: AnyStr
    password: AnyStr
    hostname: AnyStr
    port: int

class _NetlocResultMixinStr(_NetlocResultMixinBase[str], _ResultMixinStr): ...
class _NetlocResultMixinBytes(_NetlocResultMixinBase[bytes], _ResultMixinBytes): ...

class _DefragResultBase(Generic[AnyStr]):
    url: AnyStr
    fragment: AnyStr
    @overload
    def __getitem__(self, x: slice) -> AnyStr: ...
    @overload
    def __getitem__(self, x: int) -> AnyStr: ...
    def __iter__(self) -> Iterator[AnyStr]: ...

_SplitResultBase = NamedTuple(
    "_SplitResultBase",
    [
        ("scheme", str),
        ("netloc", str),
        ("path", str),
        ("query", str),
        ("fragment", str),
    ],
)
_SplitResultBytesBase = NamedTuple(
    "_SplitResultBytesBase",
    [
        ("scheme", bytes),
        ("netloc", bytes),
        ("path", bytes),
        ("query", bytes),
        ("fragment", bytes),
    ],
)

_ParseResultBase = NamedTuple(
    "_ParseResultBase",
    [
        ("scheme", str),
        ("netloc", str),
        ("path", str),
        ("params", str),
        ("query", str),
        ("fragment", str),
    ],
)
_ParseResultBytesBase = NamedTuple(
    "_ParseResultBytesBase",
    [
        ("scheme", bytes),
        ("netloc", bytes),
        ("path", bytes),
        ("params", bytes),
        ("query", bytes),
        ("fragment", bytes),
    ],
)

# Structured result objects for string data
class DefragResult(_DefragResultBase[str], _ResultMixinStr): ...
class SplitResult(_SplitResultBase, _NetlocResultMixinStr): ...
class ParseResult(_ParseResultBase, _NetlocResultMixinStr): ...

# Structured result objects for bytes data
class DefragResultBytes(_DefragResultBase[bytes], _ResultMixinBytes): ...
class SplitResultBytes(_SplitResultBytesBase, _NetlocResultMixinBytes): ...
class ParseResultBytes(_ParseResultBytesBase, _NetlocResultMixinBytes): ...

def parse_qs(
    qs: AnyStr,
    keep_blank_values: bool = ...,
    strict_parsing: bool = ...,
    encoding: str = ...,
    errors: str = ...,
) -> Dict[AnyStr, List[AnyStr]]: ...
def parse_qsl(
    qs: AnyStr,
    keep_blank_values: bool = ...,
    strict_parsing: bool = ...,
    encoding: str = ...,
    errors: str = ...,
) -> List[Tuple[AnyStr, AnyStr]]: ...
@overload
def quote(string: str, safe: _Str = ..., encoding: str = ..., errors: str = ...) -> str: ...
@overload
def quote(string: bytes, safe: _Str = ...) -> str: ...
def quote_from_bytes(bs: bytes, safe: _Str = ...) -> str: ...
@overload
def quote_plus(string: str, safe: _Str = ..., encoding: str = ..., errors: str = ...) -> str: ...
@overload
def quote_plus(string: bytes, safe: _Str = ...) -> str: ...
def unquote(string: str, encoding: str = ..., errors: str = ...) -> str: ...
def unquote_to_bytes(string: _Str) -> bytes: ...
def unquote_plus(string: str, encoding: str = ..., errors: str = ...) -> str: ...
@overload
def urldefrag(url: str) -> DefragResult: ...
@overload
def urldefrag(url: bytes) -> DefragResultBytes: ...

if sys.version_info >= (3, 5):
    def urlencode(
        query: Union[
            Mapping[Any, Any],
            Mapping[Any, Sequence[Any]],
            Sequence[Tuple[Any, Any]],
            Sequence[Tuple[Any, Sequence[Any]]],
        ],
        doseq: bool = ...,
        safe: AnyStr = ...,
        encoding: str = ...,
        errors: str = ...,
        quote_via: Callable[[str, AnyStr, str, str], str] = ...,
    ) -> str: ...

else:
    def urlencode(
        query: Union[
            Mapping[Any, Any],
            Mapping[Any, Sequence[Any]],
            Sequence[Tuple[Any, Any]],
            Sequence[Tuple[Any, Sequence[Any]]],
        ],
        doseq: bool = ...,
        safe: AnyStr = ...,
        encoding: str = ...,
        errors: str = ...,
    ) -> str: ...

def urljoin(
    base: Optional[AnyStr], url: Optional[AnyStr], allow_fragments: bool = ...
) -> AnyStr: ...
@overload
def urlparse(url: str, scheme: str = ..., allow_fragments: bool = ...) -> ParseResult: ...
@overload
def urlparse(url: bytes, scheme: bytes = ..., allow_fragments: bool = ...) -> ParseResultBytes: ...
@overload
def urlsplit(url: Optional[str], scheme: str = ..., allow_fragments: bool = ...) -> SplitResult: ...
@overload
def urlsplit(url: bytes, scheme: bytes = ..., allow_fragments: bool = ...) -> SplitResultBytes: ...
@overload
def urlunparse(components: Tuple[AnyStr, AnyStr, AnyStr, AnyStr, AnyStr, AnyStr]) -> AnyStr: ...
@overload
def urlunparse(components: Sequence[AnyStr]) -> AnyStr: ...
@overload
def urlunsplit(components: Tuple[AnyStr, AnyStr, AnyStr, AnyStr, AnyStr]) -> AnyStr: ...
@overload
def urlunsplit(components: Sequence[AnyStr]) -> AnyStr: ...



# Stubs for networkx.algorithms.assortativity.mixing (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def attribute_mixing_dict(G, attribute, nodes: Optional[Any] = ..., normalized: bool = ...): ...
def attribute_mixing_matrix(
    G,
    attribute,
    nodes: Optional[Any] = ...,
    mapping: Optional[Any] = ...,
    normalized: bool = ...,
): ...
def degree_mixing_dict(
    G,
    x: str = ...,
    y: str = ...,
    weight: Optional[Any] = ...,
    nodes: Optional[Any] = ...,
    normalized: bool = ...,
): ...
def degree_mixing_matrix(
    G,
    x: str = ...,
    y: str = ...,
    weight: Optional[Any] = ...,
    nodes: Optional[Any] = ...,
    normalized: bool = ...,
): ...
def numeric_mixing_matrix(G, attribute, nodes: Optional[Any] = ..., normalized: bool = ...): ...
def mixing_dict(xy, normalized: bool = ...): ...



# Stubs for networkx.algorithms.connectivity.disjoint_paths (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

from networkx.algorithms.flow import edmonds_karp

default_flow_func = edmonds_karp

def edge_disjoint_paths(
    G,
    s,
    t,
    flow_func: Optional[Any] = ...,
    cutoff: Optional[Any] = ...,
    auxiliary: Optional[Any] = ...,
    residual: Optional[Any] = ...,
): ...
def node_disjoint_paths(
    G,
    s,
    t,
    flow_func: Optional[Any] = ...,
    cutoff: Optional[Any] = ...,
    auxiliary: Optional[Any] = ...,
    residual: Optional[Any] = ...,
): ...
